"""
A function to test whether two fragments are isomorphic.

We will say two fragments are isomorphic iff:
 - they have the same flow and direction
 - they have the same default exchange value
 - they have the same default termination
 - they have the same number of child flows
 - each child flow in one is isomorphic to a child flow in the other
 = recurse

What a mess this is to test.
"""
from collections import defaultdict


class TreeIsomorphismException(Exception):
    pass


class FlowMismatch(TreeIsomorphismException):
    pass


class DirectionMismatch(TreeIsomorphismException):
    pass


class ExchangeValueMismatch(TreeIsomorphismException):
    pass


class TermMismatch(TreeIsomorphismException):
    pass


class ChildFlowMismatch(TreeIsomorphismException):
    pass


def isomorphic(f1, f2, scenario=None):
    if f1.flow != f2.flow:
        raise FlowMismatch('%s | %s' % (f1, f2.flow))
    if f1.direction != f2.direction:
        raise DirectionMismatch('%s | %s' % (f1, f2.direction))
    if f1.exchange_value(scenario) != f2.exchange_value(scenario):
        if not(f1.is_reference ^ f2.is_reference):
            # if both references or neither references, the exchange values should match.
            raise ExchangeValueMismatch('%s | %s' % (f1, f2.exchange_value(scenario)))
    if f1.termination(scenario) != f2.termination(scenario):
        raise TermMismatch('%s | %s' % (f1, f2.termination(scenario)))

    '''
    Create two collections of child flows and compare them pairwise and recursively.  Build a list of matched 
    child flows from f1, deplete the list of available child flows from f2.  When the built list equals the total
    set of f1 child flows and the depleted list goes to zero, the child flows are identical.
    '''
    cf1 = defaultdict(set)
    cf2 = defaultdict(set)
    c1 = set(c for c in f1.child_flows)
    c2 = set(c for c in f2.child_flows)
    for c in c1:
        cf1[c.flow, c.direction].add(c)
    for c in c2:
        cf2[c.flow, c.direction].add(c)
    if set(cf1.keys()) != set(cf2.keys()):
        raise ChildFlowMismatch('%s\n%s' % (f1, f2))
    c1check = set()
    for c in c1:
        for x in cf2[c.flow, c.direction]:
            if x not in c2:
                continue
            try:
                # this is necessary in case fragments have multiple children with the same flow + direction
                isomorphic(c, x)
            except TreeIsomorphismException:
                # only problem with this is that it conceals the exception that caused the failure
                continue
            c2.remove(x)
            c1check.add(c)

    if len(c2) == 0:
        if c1check == c1:
            return True
        else:
            raise ChildFlowMismatch('missed: %s' % ('\n'.join(str(k) for k in c1 - c1check)))
    else:
        if c1check == c1:
            raise ChildFlowMismatch('unaccounted: %s' % ('\n'.join(str(k) for k in c2)))
        else:
            c1miss = c1 - c1check
            print('Left in f1:')
            for c in c1miss:
                print(c)
            print('Left in f2:')
            for c in c2:
                print(c)
            raise TreeIsomorphismException



import logging

from antelope import InvalidQuery, EntityNotFound, ItemNotFound, QuantityRequired
from antelope_core.catalog_query import CatalogQuery
from antelope_core.contexts import NullContext
from antelope.models import Characterization as CharacterizationModel

from .interfaces.iforeground import AntelopeForegroundInterface
from .models import FragmentFlow as FragmentFlowModel, FragmentBranch as FragmentBranchModel
from .fragment_flows import FragmentFlow
from .terminations import FlowTermination


class ForegroundNotSafe(Exception):
    """
    This foreground has not been loaded yet. keep our references unresolved
    """
    pass


class FragmentBranch(object):
    def __init__(self, node, anchor, group='', scenario=None, magnitude=None, is_cutoff=None):
        self.node = node
        self.anchor = anchor
        self.group = group
        self.scenario = scenario
        self.magnitude = magnitude
        self.is_cutoff = bool(is_cutoff)

    @property
    def parent(self):
        return self.node.reference_entity

    @property
    def name(self):
        return self.anchor.name

    @property
    def unit(self):
        return self.node.flow.unit

    @property
    def is_balance_flow(self):
        return self.node.is_balance


class ForegroundQuery(CatalogQuery, AntelopeForegroundInterface):
    def cascade(self, origin):
        if origin in self._catalog.foregrounds:
            return self._catalog.foreground(origin)
        return super(ForegroundQuery, self).cascade(origin)

    def characterize(self, flowable, ref_quantity, query_quantity, value, context=None, location='GLO', **kwargs):
        """
        Now that we are back in the foreground, we want to go back to the implementation's own characterize.
        we overrule CatalogQuery.characterize() and go back to the abstract interface.
        We also want to make sure that we get rid of stale characterizations and unit scores.
        SEE? I TOLD you flows_for_flowable was important.
        {all of this goofy stuff would go away with a proper graph database, grumble}
        :param flowable:
        :param ref_quantity:
        :param query_quantity:
        :param value:
        :param context:
        :param location:
        :param kwargs:
        :return:
        """
        cf = self._perform_query('quantity', 'characterize', QuantityRequired,
                                 flowable, ref_quantity, query_quantity, value,
                                 context=context, location=location, **kwargs)
        qq = self.get_canonical(query_quantity)
        if qq.is_lcia_method:
            self._catalog.clear_unit_scores(qq)
        for flow in self._tm.flows_for_flowable(flowable):
            flow.clear_chars(qq)
        if isinstance(cf, CharacterizationModel):
            return self._resolve_cf(cf)
        else:
            return cf

    """
    Add foreground interface to query object.
    We also need to add lubricating code to translate between pydantic models and operable objects
    """
    def make_term_from_anchor(self, parent, anchor, scenario, flow_conversion=None):
        if anchor.is_null:
            term = FlowTermination.null(parent)
        else:
            if anchor.anchor_flow:
                term_flow = self.get(anchor.anchor_flow.entity_id,
                                     origin=anchor.anchor_flow.origin)
            else:
                term_flow = None

            """
            def characterize(self, flowable, ref_quantity, query_quantity, value, context=None, location='GLO', **kws):
            """
            if anchor.context:
                cx = self.get_context(anchor.context)
                if term_flow is not None and flow_conversion is not None and flow_conversion != 1.0:
                    print('Term CF %s : %s [%g]' % (parent.link, cx, flow_conversion))
                    # log reported flow conversion.  Some shit to sort out w/r/t/ context
                    # we want this characterization to apply locally
                    super(ForegroundQuery, self).characterize(parent.flow.name, parent.flow.reference_entity,
                                                              term_flow.reference_entity,
                                                              flow_conversion, context=cx)

                term = FlowTermination(parent, cx, term_flow=term_flow,
                                       descend=anchor.descend)
            elif anchor.node:
                if anchor.node.link == parent.link:  # self-termination
                    term_node = parent
                else:
                    term_node = self.get(anchor.node.entity_id, origin=anchor.node.origin)
                if flow_conversion is not None and flow_conversion != 1.0:
                    rx = term_node.reference(term_flow)
                    print('Term CF %s : %s [%g]' % (parent.link, term_node.link, flow_conversion))
                    # log reported flow conversion.  Some shit to sort out w/r/t/ context
                    super(ForegroundQuery, self).characterize(parent.flow.name, parent.flow.reference_entity,
                                                              rx.flow.reference_entity,
                                                              flow_conversion,
                                                              context=(term_node.origin, term_node.external_ref))

                term = FlowTermination(parent, term_node, term_flow=term_flow,
                                       descend=anchor.descend)
            else:
                term = FlowTermination.null(parent)
        if anchor.score_cache:
            ar = self._catalog.get_archive(self.origin)
            term._deserialize_score_cache(ar, anchor.score_cache, scenario)

        if parent.is_entity:
            logging.warning('make_term_from_anchor called by fragment entity %s' % parent.link)
        else:
            parent._anchors[scenario] = term  # just deal with it
        return term

    def _make_fragment_branch(self, n):
        if isinstance(n, FragmentBranchModel):
            frag = self.get(n.node.entity_id, origin=n.node.origin)
            term = self.make_term_from_anchor(frag, n.anchor, n.scenario, 1.0)
            return FragmentBranch(frag, term, group=n.group, scenario=n.scenario, magnitude=n.magnitude,
                                  is_cutoff=n.is_cutoff)
        return n

    def nodes(self, origin=None, **kwargs):
        ns = super(ForegroundQuery, self).nodes(origin=origin, **kwargs)
        return [self._make_fragment_branch(n) for n in ns]

    def _make_fragment_flow(self, ff_model):
        if isinstance(ff_model, FragmentFlowModel):
            frag = self.get(ff_model.fragment.entity_id, origin=ff_model.fragment.origin)

            # we have to do this manually because legacy code is terrible
            term = self.make_term_from_anchor(frag, ff_model.anchor, ff_model.anchor_scenario, ff_model.flow_conversion)

            the_ff = FragmentFlow(frag, ff_model.magnitude, ff_model.node_weight, term, ff_model.is_conserved,
                                  match_ev=ff_model.scenario, match_term=ff_model.anchor_scenario,
                                  flow_conversion=ff_model.flow_conversion)

            if len(ff_model.subfragments) > 0:
                subfrags = [self._make_fragment_flow(FragmentFlowModel(**ff)) for ff in ff_model.subfragments]
                the_ff.aggregate_subfragments(subfrags, (ff_model.anchor_scenario,))

            return the_ff

        return ff_model

    def traverse(self, fragment, scenario=None, **kwargs):
        ffs = super(ForegroundQuery, self).traverse(fragment, scenario=scenario, **kwargs)
        return [self._make_fragment_flow(ff) for ff in ffs]

    def cutoff_flows(self, fragment, scenario=None, **kwargs):
        ffs = super(ForegroundQuery, self).cutoff_flows(fragment, scenario=scenario, **kwargs)
        return [self._make_fragment_flow(ff) for ff in ffs]

    def activity(self, fragment, scenario=None, **kwargs):
        ffs = super(ForegroundQuery, self).activity(fragment, scenario=scenario, **kwargs)
        return [self._make_fragment_flow(ff) for ff in ffs]

    def fragment_lcia(self, fragment, quantity_ref, scenario=None, **kwargs):
        ress = super(ForegroundQuery, self).fragment_lcia(fragment, quantity_ref, scenario=scenario, **kwargs)
        return self._cycle_through_ress(ress, fragment, quantity_ref)

    def flowable(self, item):
        return self._tm.get_flowable(item)

    def __getitem__(self, item):
        try:
            return self.get(item)
        except EntityNotFound:
            return None  # I know, it's so bad. Plan is to break this and use downstream errors to expunge the practice


class QueryIsDelayed(InvalidQuery):
    """
    This indicates a foreground that has been queued for initialization (recursively)-- should become initialized
    before the current operation is concluded, thus allowing the DelayedQuery to function
    """
    pass


class MissingResource(InvalidQuery):
    """
    This indicates an UnknownOrigin exception was encountered when attempting to resolve a reference-- requires
    intervention from the user to supply a resource to fulfill the DelayedQuery
    """
    pass


class DelayedQuery(ForegroundQuery):
    """
    unresolved query that can sub itself in
    all it needs to do is raise a validation error until it's switched on
    """
    _home = None

    def __init__(self, origin, catalog, home, **kwargs):
        self._home = home
        super(DelayedQuery, self).__init__(origin, catalog, **kwargs)

    def get_context(self, term, **kwargs):
        cx = self._catalog.lcia_engine[term]
        if cx is None:
            return NullContext
        return cx

    def validate(self):
        if self._catalog.is_in_queue(self._home):
            return True  # this has to be true in order for the ref to operate while it is delayed
        try:
            return super(DelayedQuery, self).validate()
        except MissingResource:
            return True  # likewise

    def get_item(self, external_ref, item):
        try:
            return super(DelayedQuery, self).get_item(external_ref, item)
        except MissingResource:
            raise ItemNotFound

    def is_lcia_engine(self, **kwargs):
        try:
            return super(DelayedQuery, self).is_lcia_engine(**kwargs)
        except MissingResource:
            return False

    def _perform_query(self, itype, attrname, exc, *args, **kwargs):
        if self._catalog.is_in_queue(self._home):
            raise QueryIsDelayed(self.origin, self._home)
        return super(DelayedQuery, self)._perform_query(itype, attrname, exc, *args, **kwargs)



"""
General toolkit for comprehensive comparison of local and remote calrecycle implementations
"""

from antelope_catalog import LcCatalog
from antelope_catalog.data_sources.local import make_config
from math import isclose
import os
import json

CALRECYCLE_CAT = '/data/LCI/CalRecycle'

cat = LcCatalog(CALRECYCLE_CAT)
if 'calrecycle.antelope' not in cat.origins:
    cat.new_resource('calrecycle.antelope', 'http://www.antelope-lca.net/uo-lca/api/', 'AntelopeV1Client', store=False,
                      interfaces=['index', 'inventory', 'quantity'], quiet=True)

fg = make_config('calrecycle').foreground(cat)


def _get_local_lcia(lcia_id):
    remote_lcia = cat.query('calrecycle.antelope').get('lciamethods/%d' % lcia_id)
    return cat.query('calrecycle.lcia').get(remote_lcia.uuid)


def validate_antelope(lcia_id, **kwargs):
    lcia_m = _get_local_lcia(lcia_id)
    print('Validating for %s' % lcia_m)
    for i in range(55):
        key = 'fragments/%d' % (i+1)
        res_a = cat.query('calrecycle.antelope').get(key).fragment_lcia('lciamethods/%d' % lcia_id)
        res_l = fg[key].fragment_lcia(lcia_m, 'quell_eg')
        if isclose(res_a.total(), res_l.total(), **kwargs):
            print('valid: %s' % key)
        else:
            ratio = res_a.total() / res_l.total()
            print('\n%s : %.10g\n%s\n%s' % (key, ratio, res_a, res_l))
            yield key, ratio


def _sort_set(s):
    return sorted(s, key=lambda x: int(x.split('/')[1]))


def validate_fragments(*args, rel_tol=1e-8, save=False, **kwargs):
    """
    Supply a list of LCIA method IDs known to the remote antelope interface. This will return a list of fragments
    whose LCIA results match to the specified tolerance for all supplied methods

    if save is True, write the results to disk
    otherwise [default], if a save file exists, check to see whether the newly generated results match the saved ones.
    :param args:
    :param rel_tol: argument to isclose (also pass other kwargs)
    :param save: whether to save the results for later validation
    :return:
    """
    all_frags = set(k.external_ref for k in fg.fragments())
    result = {'methods': sorted(args), 'rel_tol': rel_tol, 'kwargs': kwargs}
    for lcia_id in args:
        fail_dict = {fail: ratio for fail, ratio in validate_antelope(lcia_id, rel_tol=rel_tol, **kwargs)}
        result['fail_%d' % lcia_id] = fail_dict
        all_frags -= set(fail_dict.keys())

    result['pass'] = _sort_set(all_frags)

    fname_suffix = '_'.join(str(k) for k in sorted(args))
    dest_file = os.path.join(os.path.dirname(__file__), 'validation_results_%s.json' % fname_suffix)

    if save:
        with open(dest_file, 'w') as fp:
            json.dump(result, fp, indent=2, sort_keys=True)
    else:
        check_result(dest_file, result)
    return result


def check_result(dest_file, result):
    passing = None
    if os.path.exists(dest_file):
        passing = True
        with open(dest_file, 'r') as fp:
            saved = json.load(fp)
        for k, v in saved.items():
            if isinstance(v, dict):
                chk = set(result[k].keys()) == set(v.keys())
                suf = ' (keys)'
            else:
                chk = result[k] == v
                suf = ''
            passing &= chk
            print('%s: %s%s' % (k, chk, suf))
    else:
        print('No saved result file %s' % dest_file)
    return passing



"""
CalRecycle LCA
Fragment Importer

The CalRecycle fragment model is stored in four files:

 Fragment.csv - (F) names and uuids for 55 fragments
 FragmentFlow.csv - (FF) data for 888 fragment flows
 FragmentNodeProcess.csv - (FNP) data for 247 nodes terminated to foreground processes
 FragmentNodeFragment.csv - (FNF) data for 76 nodes terminated to sub-fragments

The process data all come from 5 different ILCD archives:
Full UO LCA Flat Export BK
Full UO LCA Flat Export Improper
Full UO LCA Flat Export Ecoinvent
Full UO LCA Flat Export PE-SP22
Full UO LCA Flat Export PE-SP24

FNP references unique processes by uuid + version, but the version distinction is only relevant for differentiating
PE-SP22 and PE-SP24.  In the future those will be distinguished by semantic references and not UUIDs.  For the present,
the version differences can be ignored as they can be corrected later on.

The objective here is to construct the 55 fragments and replicate the LCIA results.  Well, first to construct the
fragments. And then to tune the qdb + catalog interfaces until the LCIA results correspond.

What do we need to do for that?

"""
import os
import csv
from collections import namedtuple
from antelope import comp_dir

from ..entities.fragment_editor import create_fragment, set_child_exchanges
from antelope_core.catalog.lc_resolver import ResourceNotFound

from antelope_core.data_sources.data_source import DataSource, DataCollection
from ..entities.fragments import BalanceAlreadySet


SemanticResource = namedtuple('SemanticResource', ('path', 'ref', 'privacy', 'priority'))


# DATA_ROOT = '/data/GitHub/CalRecycle/LCA_Data/'

data_main = SemanticResource('Full UO LCA Flat Export BK', 'calrecycle.uolca.core', 0, 10)
data_improper = SemanticResource('Full UO LCA Flat Export Improper', 'calrecycle.uolca.improper', 0, 20)
data_ecoinvent = SemanticResource('Full UO LCA Flat Export Ecoinvent', 'calrecycle.uolca.ecoinvent.2.2', 1, 30)
data_pe24 = SemanticResource('Full UO LCA Flat Export PE-SP24', 'calrecycle.uolca.pe.sp24', 1, 40)
data_pe22 = SemanticResource('Full UO LCA Flat Export PE-SP22', 'calrecycle.uolca.pe.sp22', 1, 30)  # default

lcia_elcd = SemanticResource('ELCD-LCIA', 'calrecycle.lcia.elcd', 0, 70)
lcia_traci = SemanticResource('TRACI Core-4 Export', 'calrecycle.lcia.traci.2.0', 0, 70)

sources = (data_main, data_improper, data_ecoinvent, data_pe22, data_pe24)
lcia_sources = (lcia_elcd, lcia_traci)

private_roots = [k.ref for k in sources if k.privacy > 0]


class CalRecycleArchive(DataSource):
    _ds_type = 'IlcdArchive'

    def __init__(self, data_root, semantic_resource):
        archive_path = os.path.join(data_root, semantic_resource.path)
        super(CalRecycleArchive, self).__init__(archive_path)
        self._info = semantic_resource

    @property
    def origins(self):
        for k in (self._info.ref,):
            yield k

    def interfaces(self, ref):
        yield 'inventory'
        if self._info.privacy > 0 or ref == 'calrecycle.uolca.core':
            yield 'background'
        if self._info.privacy == 0:
            yield 'quantity'

    def make_resources(self, ref):
        if ref == self._info.ref:
            yield self._make_resource(ref, self.root, privacy=self._info.privacy, priority=self._info.priority,
                                      interfaces=[k for k in self.interfaces(ref)])


class CalRecycleLcia(CalRecycleArchive):
    _ds_type = 'IlcdLcia'

    def interfaces(self, ref):
        yield 'quantity'


class CalRecycleConfig(DataCollection):
    def factory(self, data_root, **kwargs):
        for source in sources:
            yield CalRecycleArchive(data_root, source)
        for source in lcia_sources:
            yield CalRecycleLcia(data_root, source)

    def foreground(self, cat, **kwargs):
        self.register_all_resources(cat)
        return CalRecycleImporter.run_import(self.root, cat, **kwargs)


'''
def install_resources(cat):
    for a in sources:
        if a.privacy > 0:
            iface = ['inventory', 'background']
        else:
            iface = 'inventory'
        res = LcResource(a.root, a.path, 'IlcdArchive', interfaces=iface,
                         priority=a.priority, static=False)
        cat.add_resource(res)
    for q in (lcia_elcd, lcia_traci):
        res = LcResource(q.root, q.path, 'IlcdLcia', interfaces='quantity', priority=q.priority, static=False)
        cat.add_resource(res)
'''


def direction_map(dir_id):
    return {'1': 'Input',
            1: 'Input',
            '2': 'Output',
            2: 'Output'}[dir_id]


def read_csv(file):
    """
    Creates a dict array where the keys are given in the first row of the CSV file
    :param file:
    :return: list of dicts.
    """
    with open(file, 'r') as fp:
        reader = csv.reader(fp, delimiter=',', quotechar='"')
        headers = next(reader)
        d = []
        for row in reader:
            d.append(dict((headers[i], v) for i, v in enumerate(row)))
    return d


def dict_from_csv(file):
    """
    Creates a dict of string first-column -> second column.
    :param file:
    :return:
    """
    with open(file, 'r') as fp:
        reader = csv.reader(fp, delimiter=',', quotechar='"')
        headers = next(reader)
        if len(headers) > 2:
            raise ValueError('This function only works with 2-column tables')
        d = dict()
        for row in reader:
            d[row[0]] = row[1]
        return d


class CalRecycleImporter(object):
    """
    runs to create
    """
    @classmethod
    def run_import(cls, data_root, cat, origin='calrecycle.uolca', fg_path='fg', **kwargs):
        """

        :param data_root:
        :param cat:
        :param origin: ['calrecycle.uolca']
        :param fg_path: ['fg'] either a directory name in the catalog root, or an absolute path
        :return: a foreground archive containing the model
        """
        imp = cls(data_root, origin=origin, **kwargs)

        qi = cat.query(origin)
        try:
            fg = cat.get_archive(origin, 'foreground')
        except ResourceNotFound:
            cat.foreground(fg_path, ref=origin)  # currently this returns an interface- but op requires archive
            fg = cat.get_archive(origin, 'foreground')

        if fg.count_by_type('fragment') < len(imp.ff):
            for f in imp.ff:
                imp.fragment_from_fragment_flow(qi, f)

            imp.terminate_fragments(qi)

            imp.pe_sp24_correction(cat)

            imp.uslci_bg_elec_correction()

            imp.inv_ethylene_glycol_correction()

            imp.set_balances()

            for f in imp.fragments:
                fg.add_entity_and_children(f)
                if f.reference_entity is None:
                    f_id = imp.fragment_flow_by_index(f['FragmentFlowID'])['FragmentID']
                    fg.name_fragment(f, 'fragments/%s' % f_id)

            fg.save()

        return fg

    @property
    def fragment_dir(self):
        return os.path.join(self._root, 'fragments')

    def _print(self, *args):
        if not self._quiet:
            print(*args)

    def __init__(self, data_root, origin, quiet=True):
        if origin is None:
            raise ValueError('origin is required')
        self._origin = origin
        self._quiet = quiet
        self._root = data_root
        self._stages = dict_from_csv(os.path.join(self.fragment_dir, 'FragmentStage.csv'))
        self._f = read_csv(os.path.join(self.fragment_dir, 'Fragment.csv'))
        self._ff = read_csv(os.path.join(self.fragment_dir, 'FragmentFlowI.csv'))
        self._fnp = read_csv(os.path.join(self.fragment_dir, 'FragmentNodeProcess.csv'))
        self._fnf = read_csv(os.path.join(self.fragment_dir, 'FragmentNodeFragment.csv'))
        self._bg = read_csv(os.path.join(self.fragment_dir, 'Background.csv'))

        for term in self.fnp:
            ffid = int(term['FragmentFlowID']) - 1
            if 'Process' in self.ff[ffid]:
                raise ValueError('Fragment %d is multiply terminated' % term['FragmentFlowID'])
            self.ff[ffid]['Process'] = term
        for term in self.fnf:
            ffid = int(term['FragmentFlowID']) - 1
            if 'SubFragment' in self.ff[ffid]:
                raise ValueError('Fragment %d is multiply terminated' % term['FragmentFlowID'])
            self.ff[ffid]['SubFragment'] = term
        for frag in self.f:
            ffid = int(frag['ReferenceFragmentFlowID']) - 1
            if 'Fragment' in self.ff[ffid]:
                raise ValueError('Fragment %d is multiply referenced' % frag['ReferenceFragmentFlowID'])
            self.ff[ffid]['Fragment'] = frag

        self._frags = dict()  # indexed with FragmentFlowID.str
        self._fragments = dict()  # indexed with FragmentID.str

    def stage(self, index):
        try:
            return self._stages[str(index)]
        except KeyError:
            return ''

    @property
    def f(self):
        return self._f

    @property
    def ff(self):
        return self._ff

    @property
    def fnp(self):
        return self._fnp

    @property
    def fnf(self):
        return self._fnf

    @property
    def fragments(self):
        for f in self._frags.values():
            yield f

    def fragment_by_index(self, index):
        return self._fragments[str(int(index))]

    def fragment_flow_by_index(self, index):
        return self.ff[int(index) - 1]

    def fragment_from_fragment_flow(self, qi, ff):
        """
        :param qi: query interface for entity lookup
        :param ff: should be a decorated entry from self._ff
        :return:
        """
        if ff['FragmentFlowID'] in self._frags:
            self._print('Already loaded: %s' % ff['FragmentFlowID'])
            return
        flow = qi.get(ff['FlowUUID'])
        direction = direction_map(ff['DirectionID'])
        stage = self.stage(ff['FragmentStageID'])
        name = ff['Name']

        # create the fragment, with or without a parent
        if ff['ParentFragmentFlowID'] == '':
            # must be a reference flow
            frag_uuid = ff['Fragment']['FragmentUUID']
            frag = create_fragment(flow, comp_dir(direction), uuid=frag_uuid, StageName=stage, Name=name,
                                   FragmentFlowID=ff['FragmentFlowID'], origin=self._origin)
            self._fragments[ff['Fragment']['FragmentID']] = frag
        else:
            try:
                parent = self._frags[ff['ParentFragmentFlowID']]
            except KeyError:
                self._print('Recursing to %s' % ff['ParentFragmentFlowID'])
                self.fragment_from_fragment_flow(qi, self.fragment_flow_by_index(ff['ParentFragmentFlowID']))
                parent = self._frags[ff['ParentFragmentFlowID']]

            frag_uuid = ff['FragmentUUID']
            frag = create_fragment(flow, direction, uuid=frag_uuid, Name=name, StageName=stage, parent=parent,
                                   FragmentFlowID=ff['FragmentFlowID'])

        # save the fragment
        if frag is None:
            raise TypeError
        self._frags[ff['FragmentFlowID']] = frag

    def terminate_fragments(self, qi, frags=None):
        for ff in self.ff:
            debug = False
            if frags is not None:
                if ff['FragmentFlowID'] not in frags:
                    continue
                self._print('FragmentFlowID: %s' % ff['FragmentFlowID'])
                debug = True

            try:
                self._terminate_fragment(qi, ff, debug=debug)
            except ZeroDivisionError:
                print('!! Skipping Frag: %s' % ff['FragmentFlowID'])

    def _terminate_fragment(self, qi, ff, debug=False):
        frag = self._frags[ff['FragmentFlowID']]
        if debug:
            frag.set_debug_threshold(999)

        flow = frag.flow
        # terminate the fragment
        if ff['NodeTypeID'] == '1':
            term_node = qi.get(ff['Process']['ProcessUUID'])
            term_flow = qi.get(ff['Process']['FlowUUID'])
            frag.clear_termination()
            frag.terminate(term_node, term_flow=term_flow)
            if term_node.origin in private_roots:
                # frag.set_background()
                pass
            else:
                set_child_exchanges(frag)
        elif ff['NodeTypeID'] == '2':
            term_frag = self._fragments[ff['SubFragment']['SubFragmentID']]
            term_flow = qi.get(ff['SubFragment']['FlowUUID'])
            if term_flow is term_frag.flow:
                term_node = term_frag
            else:
                term_node = None
                for c in term_frag.top().traverse(None):
                    if c.term.is_null and c.fragment.flow is term_flow:
                        term_node = c.fragment
                if term_node is None:
                    raise ValueError('Could not find flow from inverse traversal')
            descend = {'1': True, '0': False}[ff['SubFragment']['Descend']]
            frag.clear_termination()
            frag.terminate(term_node, term_flow=term_flow, descend=descend)

        elif ff['NodeTypeID'] == '4':
            bg = next(row for row in self._bg
                      if row['FlowUUID'] == flow.uuid and row['DirectionID'] == ff['DirectionID'])
            if bg['TargetUUID'] != '':
                if bg['NodeTypeID'] == '1':
                    term_node = qi.get(bg['TargetUUID'])
                    if term_node.origin in private_roots:
                        pass
                        # frag.set_background()
                else:
                    term_node = next(_f for _f in self._fragments.values() if _f.uuid == bg['TargetUUID'])
                frag.clear_termination()
                frag.terminate(term_node, term_flow=flow, descend=False)
        return frag

    def set_balances(self):
        for f in self.ff:
            if f['NodeTypeID'] == '1':
                if f['Process']['ConservationFFID'] != '':
                    try:
                        self._frags[f['Process']['ConservationFFID']].set_balance_flow()
                    except BalanceAlreadySet:
                        pass

    '''
    Data Corrections
    '''
    def pe_sp24_correction(self, cat):
        """
        Add an 'sp24' scenario that sets all PE processes to their SP-24 versions
        :param cat:
        :return:
        """
        sp24 = cat.query('calrecycle.uolca.pe.sp24')
        for f in self._frags.values():
            if f.term.is_process:
                if f.term.term_node.origin == 'calrecycle.uolca.pe.sp22':
                    f.terminate(sp24.get(f.term.term_node.external_ref), scenario='sp24',
                                term_flow=f.term.term_flow,
                                direction=f.term.direction,
                                descend=f.term.descend)

    def uslci_bg_elec_correction(self):
        """
        Set the rolled-up USLCI processes to be computed via the background interface, to keep their
        flows out of foreground disclosures
        :return:
        """
        uslci_bg_ffids = ('14', '26', '38', '50', '62', '74')
        for i in uslci_bg_ffids:
            frag = self._frags[i]
            # frag.set_background()

    def inv_ethylene_glycol_correction(self):
        """
        Legacy Antelope online tool fails to generate nonzero impact scores for inverted processes because
        the directions don't match on the join.  The model has only one inverted process: avoided ethylene glycol,
        thinkstep process with UUID 'df09efef-4d73-4b45-a899-1c6d1ca97da0'
        :return:
        """
        for f in self._frags.values():
            if f.term.is_process:
                if f.term.term_node.uuid == 'df09efef-4d73-4b45-a899-1c6d1ca97da0':
                    f.set_exchange_value(scenario='quell_eg', value=0.0)


if __name__ == '__main__':
    print('foo')



from antelope_core.implementations import IndexImplementation


class AntelopeIndexImplementation(IndexImplementation):
    def re_index(self, cutoffs=False):
        self._archive.ti = {}

    def targets(self, flow_ref, direction=None, **kwargs):
        raise NotImplementedError

    def count(self, entity_type, **kwargs):
        raise NotImplementedError



"""
Client for the oryx Foreground server

This is to be the same as the XdbServer, just with different methods defined
"""
import logging

from antelope import UnknownOrigin, comp_dir
from antelope_core.providers.xdb_client import XdbClient, _ref
from antelope_core.providers.xdb_client.xdb_entities import XdbEntity
from antelope_core.implementations import BasicImplementation
from antelope.models import OriginCount, LciaResult as LciaResultModel, EntityRef, FlowEntity

from ..interfaces import AntelopeForegroundInterface
from ..refs.fragment_ref import FragmentRef, ParentFragment
from .lc_foreground import AmbiguousReference, FragmentNotFound

from ..models import (LcForeground, FragmentFlow, FragmentRef as FragmentRefModel, MissingResource,
                      FragmentBranch, FragmentEntity, Anchor, ForegroundRelease, Observation)

from requests.exceptions import HTTPError
from uuid import uuid4


class MalformedOryxEntity(Exception):
    """
    something is wrong with the entity model
    """
    pass


class OryxEntity(XdbEntity):
    @property
    def uuid(self):
        if hasattr(self._model, 'entity_uuid'):
            return self._model.entity_uuid
        return None

    def make_ref(self, query):
        if self._ref is not None:
            return self._ref

        if self.entity_type == 'fragment':
            """
            This is complicated because there are a couple different possibilities for the model type.
            If the model is a FragmentEntity, then it contains a 'flow' attribute which is actually a FlowEntity,
            but if the model is a FragmentRef, then its flow and direction are stored along with other 
            entity properties, and they will not be converted into pydantic types but kept as dicts
            
            we also need to handle parents, with reference fragments (having None parents) being replaced in the API
            layer with 404 errors and thereby getting caught and replaced with ParentFragment exceptions... 
            
            and also UUIDs which arrive from Entity models as properties and from FragmentRef models as entity_uuid
            attributes
            """
            args = {k: v for k, v in self._model.properties.items()}
            f = args.pop('flow', None)
            d = args.pop('direction', None)
            if hasattr(self._model, 'is_balance_flow'):
                args['balance_flow'] = self._model.is_balance_flow

            parent = args.pop('parent', ParentFragment) or ParentFragment
            if hasattr(self._model, 'flow'):
                the_origin = self._model.flow.origin
                the_id = self._model.flow.entity_id
                direction = self._model.direction
                args['uuid'] = self._model.entity_uuid
            else:
                if f is None:
                    print(self._model.model_dump_json(indent=2))
                    raise MalformedOryxEntity(self.link)
                the_origin = f['origin']
                the_id = f['entity_id']
                direction = d

            if hasattr(self._model, 'parent'):
                parent = self._model.parent or ParentFragment

            try:
                flow = query.cascade(the_origin).get(the_id)  # get locally
            except UnknownOrigin:
                flow = query.get(the_id, origin=the_origin)  # get remotely

            if self.origin != query.origin:
                args['masquerade'] = self.origin

            if hasattr(self._model, 'exchange_values'):
                args['exchange_values'] = self._model.exchange_values

            ref = FragmentRef(self.external_ref, query,
                              flow=flow, direction=direction, parent=parent, **args)

            if hasattr(self._model, 'anchors'):
                ref.anchors(**self._model.anchors)

            self._ref = ref
            return ref

        return super(OryxEntity, self).make_ref(query)


class OryxClient(XdbClient):

    _base_type = OryxEntity

    @property
    def missing(self):
        return self.r.origin_get_many(MissingResource, 'missing')

    def clear_unit_scores(self, lcia_method):
        pass

    def __init__(self, *args, catalog=None, **kwargs):
        """
        Not sure we need the catalog yet, but LcResource gives it to us, so let's hold on to it
        :param args:
        :param catalog:
        :param kwargs:
        """
        self._catalog = catalog
        super(OryxClient, self).__init__(*args, **kwargs)

    @property
    def query(self):
        return self._catalog.query(self.ref)

    def make_interface(self, iface):
        if iface == 'foreground':
            return OryxFgImplementation(self)
        return super(OryxClient, self).make_interface(iface)

    def _model_to_entity(self, model):
        if model.entity_type == 'fragment':
            ''' check for flow spec '''
            if hasattr(model, 'flow'):
                self.get_or_make(model.flow)
            else:
                if hasattr(model, 'properties') and 'flow' in model.properties:
                    self.get_or_make(FlowEntity(**model.properties['flow']))
                else:
                    model = self._requester.origin_get_one(FragmentEntity, model.origin, 'fragments',
                                                           model.entity_id)
                    self.get_or_make(model.flow)

        ent = super(OryxClient, self)._model_to_entity(model)
        if ent.uuid is not None:
            self._entities[ent.uuid] = ent
        return ent

    def frag(self, string, many=True, strict=False):
        found = None
        for v in self.entities_by_type('fragment'):
            if v.uuid.startswith(string):
                if many or not strict:
                    return v
                else:
                    if found is None:
                        found = v
                    else:
                        raise AmbiguousReference(string)
        if found is None:
            raise FragmentNotFound(string)
        return found


class OryxFgImplementation(BasicImplementation, AntelopeForegroundInterface):
    """
    We don't need to REimplement anything in XdbClient because oryx server should behave the same to the same routes
    (but that means we need to reimplement everything in OryxServer)
    """
    def _o(self, obj=None):
        """
        Key difference between the Xdb implementation is: the xdb implementation is strongly tied to its origin,
        but the foreground can refer to entities with various origins.

        To handle this, we *masquerade* the query (to the primary origin) with the entity's authentic origin (just as
        we do with local.qdb). this happens automatically in entity.make_ref() when the query origin doesn't match the
        entity origin

        then in our requester we unset_origin() and issue origin, ref explicitly.

        _o is the mechanism for this.

        Implies that client code is expected to supply a true entity and not a string ref-- this is potentially a
        problem

        returns either the object's origin, if it is an object, or the archive's ref

        :param obj:
        :return:
        """
        if hasattr(obj, 'origin'):
            return obj.origin
        return self._archive.ref

    @property
    def delayed(self):
        return self._archive.delayed

    @property
    def unresolved(self):
        return self._archive.unresolved

    def get(self, external_ref, **kwargs):
        print('I have a theory this tranche of code never gets run %s' % external_ref)
        return self._archive.query.get(external_ref, **kwargs)

    # foreground resource operations-- non-masqueraded
    def fragments(self, **kwargs):
        llargs = {k.lower(): v for k, v in kwargs.items()}
        return [self._archive.get_or_make(k) for k in self._archive.r.get_many(FragmentRefModel, 'fragments', **llargs)]

    def frag(self, string, many=False, **kwargs):
        """
        this just runs locally
        :param string:
        :param many:
        :param kwargs:
        :return:
        """
        return self._archive.frag(string, many=many, **kwargs)

    def frags(self, string, **kwargs):
        return [self._archive.get_or_make(k) for k in self._archive.r.get_many(FragmentRefModel,
                                                                               'frags', string)]

    def knobs(self, search=None, reference=False):
        return [self._archive.get_or_make(k) for k in self._archive.r.get_many(FragmentRefModel,
                                                                               'knobs', search=search,
                                                                               reference=reference)]

    @staticmethod
    def _sc(scenario):
        if scenario is None:
            return None
        elif isinstance(scenario, int):
            return str(scenario)
        elif isinstance(scenario, str):
            return scenario
        elif hasattr(scenario, '__iter__'):
            return ','.join(map(str, scenario))
        raise TypeError(scenario)

    def ev(self, frag, scenario=None, observed=None):
        scenario = self._sc(scenario)
        if scenario is None:
            return self._archive.r.origin_get_one(float, self._o(frag), 'fragments', _ref(frag), 'ev',
                                                  observed=bool(observed))
        return self._archive.r.origin_get_one(float, self._o(frag), 'fragments', _ref(frag), 'ev',
                                              scenario=scenario)

    def new_fragment(self, flow, direction, parent=None, external_ref=None, uuid=None, balance=None,
                     value=None, exchange_value=None, units=None,
                     **kwargs):
        """
        The strategy here is to just POST a fragment entity.
        :param flow:
        :param direction:
        :param parent:
        :param external_ref:
        :param uuid:
        :param balance:
        :param value:
        :param exchange_value: synonym for value
        :param units:
        :param kwargs:
        :return:
        """
        if uuid is None:
            uuid = str(uuid4())
        if external_ref is None:
            entity_id = uuid
        else:
            entity_id = external_ref
        if parent is None:
            direction = comp_dir(direction)
            p = None
            origin = self.origin
        else:
            p = _ref(parent)
            origin = self._o(parent)
        frag = FragmentEntity(origin=origin, entity_id=entity_id, flow=FlowEntity.from_flow(flow), direction=direction,
                              entity_uuid=uuid, is_balance_flow=balance, exchange_values={'0': 1.0}, anchors=dict(),
                              parent=p, properties=kwargs)
        fragments = [self._archive.get_or_make(k) for k in
                     self._archive.r.post_return_many([frag.model_dump()], FragmentRefModel, 'fragments')]
        if len(fragments) > 1:
            logging.warning('Multiple fragments returned!')
        fragment = fragments[0]
        if exchange_value is None:
            if value is not None:
                exchange_value = value
        if exchange_value:
            self.observe(fragment, exchange_value=exchange_value, units=units)
        return fragment

    def post_foreground(self, fg, save_unit_scores=False):
        pydantic_fg = LcForeground.from_foreground_archive(fg.archive, save_unit_scores=save_unit_scores)
        return self._archive.r.post_return_one(pydantic_fg.dict(), OriginCount, 'post_foreground')

    def post_entity_refs(self, entities, **kwargs):
        post_ents = [p if isinstance(p, EntityRef) else EntityRef.from_entity(p) for p in entities]
        return self._archive.r.post_return_one([p.model_dump() for p in post_ents], OriginCount, 'entity_refs')

    def save(self, description=None, author=None, notes=None, major=False):
        if description is None:
            raise ValueError('Foreground release description must be provided')
        release = ForegroundRelease(major=major, description=description, author=author, notes=notes)
        return self._archive.r.post_return_one(release.model_dump(), bool, 'save_foreground')

    def restore(self):
        return self._archive.r.post_return_one(None, bool, 'restore_foreground')

    # Entity operations- masqueraded
    def get_reference(self, key):
        try:
            # !TODO! key will always be an external_ref so _o(key) will fail
            parent = self._archive.r.origin_get_one(FragmentRefModel, self._o(key), _ref(key), 'reference')
        except HTTPError as e:
            if e.args[0] == 400:
                raise ParentFragment
            raise e
        return self._archive.get_or_make(parent)

    def get_fragment(self, fragment):
        """
        detailed version of a fragment
        :param fragment:
        :return:
        """
        return self._archive.r.origin_get_one(FragmentEntity, self._o(fragment), 'fragments', _ref(fragment))

    def child_flows(self, fragment, **kwargs):
        return [self._archive.get_or_make(k) for k in self._archive.r.origin_get_many(FragmentRefModel,
                                                                                      self._o(fragment),
                                                                                      'fragments', _ref(fragment),
                                                                                      'child_flows')]

    def top(self, fragment, **kwargs):
        return self._archive.get_or_make(self._archive.r.origin_get_one(FragmentRefModel,
                                                                        self._o(fragment), _ref(fragment), 'top'))

    def anchors(self, fragment, **kwargs):
        a = self._archive.r.origin_get_one(dict, self._o(fragment), 'fragments', _ref(fragment), 'anchors')
        return {k: Anchor(**v) for k, v in a.items()}

    def scenarios(self, fragment, **kwargs):
        return self._archive.r.origin_get_many(str, self._o(fragment), _ref(fragment),
                                               'scenarios', **kwargs)

    def nodes(self, **kwargs):
        fbs = self._archive.r.get_many(FragmentBranch, 'nodes', **kwargs)
        for fb in fbs:
            self._archive.get_or_make(fb.node)
        return fbs

    def _get_or_make_fragment_flows(self, ffs):
        """
        whoo-ee we love danger
        :param ffs:
        :return:
        """
        for ff in ffs:
            subfrags = [FragmentFlow(**f) for f in ff.subfragments]
            self._get_or_make_fragment_flows(subfrags)
            self._archive.get_or_make(ff.node)

    def traverse(self, fragment, scenario=None, **kwargs):
        scenario = self._sc(scenario)
        ffs = self._archive.r.origin_get_many(FragmentFlow, self._o(fragment), _ref(fragment),
                                              'traverse', scenario=scenario, **kwargs)

        self._get_or_make_fragment_flows(ffs)

        return ffs

    def cutoff_flows(self, fragment, scenario=None, **kwargs):
        scenario = self._sc(scenario)
        ffs = self._archive.r.origin_get_many(FragmentFlow, self._o(fragment), _ref(fragment),
                                              'cutoff_flows', scenario=scenario, **kwargs)

        self._get_or_make_fragment_flows(ffs)

        return ffs

    def activity(self, fragment, scenario=None, **kwargs):
        scenario = self._sc(scenario)
        return self._archive.r.origin_get_many(FragmentFlow, self._o(fragment), _ref(fragment),
                                               'activity', scenario=scenario, **kwargs)

    def tree(self, fragment, scenario=None, **kwargs):
        scenario = self._sc(scenario)
        return self._archive.r.origin_get_many(FragmentBranch, self._o(fragment), _ref(fragment),
                                               'tree', scenario=scenario, **kwargs)

    def fragment_lcia(self, fragment, quantity_ref, scenario=None, mode=None, **kwargs):
        scenario = self._sc(scenario)
        if mode == 'detailed':
            return self.detailed_lcia(fragment, quantity_ref, scenario=scenario, **kwargs)
        elif mode == 'flat':
            return self.flat_lcia(fragment, quantity_ref, scenario=scenario, **kwargs)
        elif mode == 'stage':
            return self.stage_lcia(fragment, quantity_ref, scenario=scenario, **kwargs)
        elif mode == 'anchor':
            return self.anchor_lcia(fragment, quantity_ref, scenario=scenario, **kwargs)
        return self._archive.r.origin_get_one(LciaResultModel, self._o(fragment), 'fragments', _ref(fragment),
                                              'fragment_lcia',
                                              _ref(quantity_ref), scenario=scenario, **kwargs)

    def detailed_lcia(self, fragment, quantity_ref, scenario=None, **kwargs):
        scenario = self._sc(scenario)
        return self._archive.r.origin_get_one(LciaResultModel, self._o(fragment), 'fragments', _ref(fragment),
                                              'detailed_lcia',
                                              _ref(quantity_ref), scenario=scenario, **kwargs)

    def flat_lcia(self, fragment, quantity_ref, scenario=None, **kwargs):
        scenario = self._sc(scenario)
        return self._archive.r.origin_get_one(LciaResultModel, self._o(fragment), 'fragments', _ref(fragment),
                                              'lcia',
                                              _ref(quantity_ref), scenario=scenario, **kwargs)

    def stage_lcia(self, fragment, quantity_ref, scenario=None, **kwargs):
        scenario = self._sc(scenario)
        return self._archive.r.origin_get_one(LciaResultModel, self._o(fragment), 'fragments', _ref(fragment),
                                              'stage_lcia',
                                              _ref(quantity_ref), scenario=scenario, **kwargs)

    def anchor_lcia(self, fragment, quantity_ref, scenario=None, **kwargs):
        scenario = self._sc(scenario)
        return self._archive.r.origin_get_one(LciaResultModel, self._o(fragment), 'fragments', _ref(fragment),
                                              'anchor_lcia',
                                              _ref(quantity_ref), scenario=scenario, **kwargs)

    def observe(self, fragment, scenario=None, exchange_value=None, units=None, name=None, anchor=None,
                anchor_node=None, anchor_flow=None, descend=None, **kwargs):
        if anchor is None:
            if anchor_node:
                if anchor_flow:
                    anchor_flow = EntityRef.from_entity(anchor_flow)
                if anchor_node.entity_type == 'context':
                    anchor = Anchor(context=anchor_node.as_list(), anchor_flow=anchor_flow)
                else:
                    anchor = Anchor(node=EntityRef.from_entity(anchor_node), anchor_flow=anchor_flow, descend=descend)
        obs = Observation(fragment=EntityRef.from_entity(fragment),
                          scenario=scenario,
                          exchange_value=exchange_value,
                          units=units,
                          name=name,
                          anchor=anchor)
        obs_apply = self._archive.r.origin_post_return_many('observation', obs.model_dump(), Observation, **kwargs)
        # we somehow need to update the local copies of the fragment with the observations
        for o in obs_apply:
            if o.name:
                self._archive.name_fragment(o.fragment, o.name, **kwargs)
            if o.anchor:
                fragment._query.make_term_from_anchor(fragment, o.anchor, o.scenario)
        return obs_apply



"""
Defining pydantic models for foreground types- in preparation for writing the oryx API (aka. Antelope v2 foreground)

We need:

 Fragment(Entity) <- full record
 FragmentRef(EntityRef) <- minimal (but not even a name?
 FragmentLink (for tree)
 FragmentFlow (for traversal)

and that's probably it

"""
from typing import Dict, Optional, List, Set
from antelope.models import ResponseModel, EntityRef, Entity, FlowEntity, LciaResult
from antelope.xdb_tokens import ResourceSpec


UNRESOLVED_ANCHOR_TYPE = 'term'  # this is used when an anchor node's origin cannot be resolved


class MissingResource(ResponseModel):
    origin: str
    interface: str


class Anchor(ResponseModel):
    """
    An anchor is either: a terminal node designation (i.e. origin + ref) or a context, and a descent marker.
    and cached LCIA scores

    Use FlowTermination.to_anchor(term, ..) to produce
    """
    node: Optional[EntityRef] = None
    anchor_flow: Optional[EntityRef] = None
    context: Optional[List[str]] = None
    descend: bool = False
    score_cache: Optional[Dict[str, float]] = None

    @classmethod
    def from_rx(cls, rx, descend=False):
        return cls(node=EntityRef.from_entity(rx.process), anchor_flow=EntityRef.from_entity(rx.flow), descend=descend)

    def __str__(self):
        """
        Replicate FlowTermination

        :return:
          '---:' = fragment I/O
          '-O  ' = foreground node
          '-*  ' = process
          '-#  ' - sub-fragment (aggregate)
          '-#::' - sub-fragment (descend)
          '-B ' - terminated background
          '--C ' - cut-off background
          '--? ' - ungrounded catalog ref

        :return:
        """
        if self.node:
            if self.node.entity_type == 'process':
                return '-*  '
            elif self.node.entity_type == UNRESOLVED_ANCHOR_TYPE:
                return '--? '
            else:
                if self.descend:
                    return '-#::'
                else:
                    return '-#  '
        elif self.context:
            if self.context == ['None']:
                return '-)  '
            else:
                return '-== '
        else:
            return '---:'

    @property
    def type(self):
        if self.node:
            return 'node'
        elif self.context:
            return 'context'
        else:
            return 'cutoff'

    @property
    def unit(self):
        if self.is_null:
            return '--'
        if self.node:
            if self.node.entity_type == 'fragment':
                return '####'
            return '****'
        return '::::'

    @property
    def is_null(self):
        return not bool(self.node or self.context)

    @classmethod
    def null(cls):
        return cls(descend=True)

    def _term_flow_block(self):
        if self.anchor_flow:
            if self.anchor_flow.origin == self.node.origin:
                return self.anchor_flow.entity_id
            else:
                return {
                    'origin': self.anchor_flow.origin,
                    'externalId': self.anchor_flow.entity_id
                }

    def serialize(self):
        """
        emulates FlowTermination.serialize()
        :return:
        """
        if self.context:
            j = {
                'origin': 'foreground',
                'context': self.context[-1]
            }
        else:
            j = {
                'origin': self.node.origin,
                'externalId': self.node.entity_id
            }
        j['descend'] = self.descend
        tfb = self._term_flow_block()
        if tfb:
            j['termFlow'] = tfb
        if self.score_cache:
            j['scoreCache'] = self.score_cache
        return j

    def masquerade(self, masq):
        if self.node:
            if self.node.origin in masq:
                self.node.origin = masq[self.node.origin]
        if self.anchor_flow:
            if self.anchor_flow.origin in masq:
                self.anchor_flow.origin = masq[self.anchor_flow.origin]


class Anchors(ResponseModel):
    anchors: Dict[str, Anchor]

    @classmethod
    def from_fragment(cls, fragment):
        def _default_or(scenario):
            if scenario is None:
                return 'default'
            return str(scenario)

        return cls(anchors={_default_or(k): v.to_anchor(save_unit_scores=False) for k, v in fragment.terminations()})

    def masquerade(self, masq):
        for a in self.anchors.values():
            a.masquerade(masq)


class FragmentRef(Entity):
    """
    From EntityRef, inherits: origin, entity_id (<-external_ref), entity_type, properties
    we ensure put a "name" field because we want users to be able to see what they're getting
    """
    entity_type: str = 'fragment'
    flow: FlowEntity
    direction: str

    is_balance_flow: bool = False

    entity_uuid: str
    parent: Optional[str] = None

    @property
    def dirn(self):
        return {
            'Input': '-<-',
            'Output': '=>='
        }[self.direction]

    @classmethod
    def from_fragment(cls, fragment, **kwargs):
        """

        :param fragment:
        :param kwargs:
        :return:
        """
        ''' # we don't want this
        if fragment.reference_entity is None:
            dirn = comp_dir(fragment.direction)
        else:
            dirn = fragment.direction
        '''
        dirn = fragment.direction
        if fragment.parent is None:
            parent = None
        else:
            parent = fragment.parent.external_ref

        obj = cls(origin=fragment.origin, entity_id=fragment.external_ref, entity_uuid=fragment.uuid,
                  is_balance_flow=fragment.is_balance,
                  flow=FlowEntity.from_flow(fragment.flow), direction=dirn, parent=parent, properties=dict())
        obj.properties['name'] = fragment['name']

        for key, val in kwargs.items():
            try:
                obj.properties[key] = fragment[key]
            except KeyError:
                pass
        return obj

    def masquerade(self, masq):
        if self.origin in masq:
            self.origin = masq[self.origin]
        if self.flow.origin in masq:
            self.flow.origin = masq[self.flow.origin]


def _to_key(thing):
    if isinstance(thing, str):
        return thing
    elif hasattr(thing, 'uuid'):
        return thing.uuid
    elif hasattr(thing, 'link'):
        return thing.link
    raise AttributeError(thing)


def _to_set(thing_or_things):
    if thing_or_things is None:
        return set()
    try:
        return {_to_key(thing_or_things)}
    except AttributeError:
        return {_to_key(k) for k in thing_or_things}


class DescendSpec(ResponseModel):
    """
    A specification for which fragments to apply descend settings during aggregation
    """
    descend: Set = set()
    nondescend: Set = set()
    descend_all: Optional[bool] = None

    def __init__(self, descend=None, nondescend=None, descend_all=None):
        super(DescendSpec, self).__init__(descend=_to_set(descend), nondescend=_to_set(nondescend),
                                          descend_all=descend_all)

    def descend_ff(self, ff):
        """
        If the fragment is on the explicit descend or nondescend list, apply that.
        otherwise, if descend_all is True or False, apply that.
        otherwise, honor the anchor's descend setting.
        :param ff: a FragmentFlow
        :return:
        """
        if ff.term.is_null:
            return None
        ff_keys = {ff.fragment.uuid, ff.fragment.link, ff.fragment.external_ref}
        if ff_keys & self.descend:
            return True
        if ff_keys & self.nondescend:
            return False
        if self.descend_all:
            return True
        if self.descend_all is False:
            return False
        return ff.term.descend


class FragmentEntity(Entity):
    """
    From Entity, inherits: origin, entity_id (<- external_ref), entity_type, properties
    """
    entity_type: str = 'fragment'
    flow: FlowEntity  # should be full entity or just a ref? this is a full return, it should be full
    direction: str
    parent: Optional[str]
    is_balance_flow: bool = False

    entity_uuid: str  # we need uuid for consistency since we are running the same LcForeground on the backend

    exchange_values: Dict[str, float]

    anchors: Dict[str, Anchor]

    @classmethod
    def from_entity(cls, fragment, save_unit_scores=False, **kwargs):
        """
        :param fragment:
        :param save_unit_scores:
        :param kwargs:
        :return:
        """
        ''' # we don't want this
        if fragment.reference_entity is None:
            dirn = comp_dir(fragment.direction)
        else:
            dirn = fragment.direction
        '''
        dirn = fragment.direction
        j = fragment.serialize(**kwargs)
        evs = j.pop('exchangeValues')
        evs['cached'] = evs.pop('0')
        evs['observed'] = evs.pop('1')
        terms = {}
        for k, v in fragment.terminations():
            if k is None:
                k = 'default'
            a = v.to_anchor(save_unit_scores=save_unit_scores)
            terms[k] = a
        return cls(origin=fragment.origin, entity_id=fragment.external_ref, properties=j.pop('tags'),
                   entity_uuid=fragment.uuid,
                   flow=FlowEntity.from_flow(fragment.flow), direction=dirn,
                   parent=j.pop('parent'), is_balance_flow=j.pop('isBalanceFlow'),
                   exchange_values=evs, anchors=terms)

    def _serialize_evs(self):
        d = dict(**self.exchange_values)
        d["0"] = d.pop('cached')
        d["1"] = d.pop('observed')
        return d

    def _serialize_terms(self):
        terms = dict()
        for k, v in self.anchors.items():
            if v.is_null:
                terms[k] = {}
            else:
                terms[k] = v.serialize()
        return terms

    def serialize(self):
        """
        This emulates LcFragment.serialize()
        :return:
        """
        j = {
            'entityType': self.entity_type,
            'externalId': self.entity_id,
            'entityId': self.entity_uuid,
            'parent': self.parent
        }
        if self.flow.origin == self.origin:
            j['flow'] = self.flow.entity_id
        else:
            j['flow'] = '%s/%s' % (self.flow.origin, self.flow.entity_id)
        j['direction'] = self.direction
        j['isPrivate'] = False
        j['isBalanceFlow'] = self.is_balance_flow
        j['exchangeValues'] = self._serialize_evs()
        j['terminations'] = self._serialize_terms()
        j['tags'] = dict(**self.properties)
        return j

    def masquerade(self, masq):
        # I wonder if there is a better way to do this-- surely with a decorator of some sort
        if self.origin in masq:
            self.origin = masq[self.origin]
        if self.flow.origin in masq:
            self.flow.origin = masq[self.flow.origin]
        for a in self.anchors.values():
            a.masquerade(masq)


class FragmentBranch(ResponseModel):
    """
    Used to construct a tree diagram. Reports exchange values only, can be null (if not observable), and no node weights
    (because no traversal)
    """
    parent: Optional[str]
    node: FragmentRef
    name: str
    group: str  # this is the StageName, used for aggregation.. the backend must set / user specify / modeler constrain
    magnitude: Optional[float]
    unit: str
    scenario: Optional[str]
    anchor: Optional[Anchor]
    is_cutoff: bool
    is_balance_flow: bool

    @property
    def term(self):
        if self.anchor is None:
            return Anchor.null()
        return self.anchor

    @property
    def fragment(self):
        return self.node

    @property
    def term_str(self):
        if self.anchor and self.anchor.node:
            if self.node.entity_id == self.anchor.node.entity_id:
                return '-O  '
            return str(self.anchor)
        return '---:'

    @classmethod
    def from_fragment(cls, fragment, scenario=None, observed=False, group='StageName', save_unit_scores=False):
        """

        :param fragment:
        :param scenario:
        :param observed:
        :param group:
        :param save_unit_scores: score_cache must generally not be returned unless some aggregation condition is met
        :return:
        """
        if fragment.observable(scenario):
            mag = fragment.exchange_value(scenario, observed)
        else:
            mag = None
        if fragment.is_balance:
            print(' ## Balance Flow ## %s' % fragment)
        if fragment.reference_entity is None:
            parent = None
        else:
            parent = fragment.reference_entity.external_ref
        term = fragment.termination(scenario)
        anchor = term.to_anchor(save_unit_scores=save_unit_scores)
        if anchor.is_null and len(list(fragment.child_flows)) == 0:
            cutoff = True
        else:
            cutoff = False
        return cls(parent=parent, node=FragmentRef.from_fragment(fragment), name=term.name,
                   group=fragment.get(group, ''), scenario=scenario,
                   magnitude=mag, unit=fragment.flow.unit, is_balance_flow=fragment.is_balance,
                   anchor=anchor, is_cutoff=cutoff)

    @classmethod
    def from_fragment_branch(cls, n):
        if n.parent is None:
            parent = None
        else:
            parent = n.parent.external_ref
        return cls(parent=parent, node=FragmentRef.from_fragment(n.node), name=n.name,
                   group=n.group, scenario=n.scenario, magnitude=n.magnitude, unit=n.unit,
                   anchor=n.anchor.to_anchor(),
                   is_balance_flow=n.is_balance_flow, is_cutoff=n.is_cutoff)

    def masquerade(self, masq):
        self.node.masquerade(masq)
        if self.anchor:
            self.anchor.masquerade(masq)


class FragmentFlow(FragmentBranch):
    """
    A FragmentFlow is a record of a link in a traversal. Current controversy: when a flow magnitude or anchor is
    determined by a scenario specification, the FragmentFlow needs to report that.

    Also: whether foreground nodes (anchor = self) actually have anchors (decision: no, the 'foreground is self'
    designation is actually a hack and is redundant. the relevant trait is whether it has child flows, which can be
    known by the constructor)

    hmm except all FragmentFlows have anchors. I think we need to preserve the anchor-is-self, and simply test for it
    when doing operations.

    """
    magnitude: float
    flow_conversion: float
    scenario: Optional[str]
    node_weight: float
    anchor_scenario: Optional[str]
    is_conserved: bool
    subfragments: List = []

    def __str__(self):
        """
        Replicate the FragmentFlow behavior
                return '%.5s  %10.3g [%6s] %s %s' % (self.fragment.uuid, self.magnitude, self.fragment.direction,
                                             self.term, self.name)

        :return:
        """
        if self.anchor is None:
            anc = '---:'
        else:
            anc = str(self.anchor)
            if self.anchor.node:
                if self.anchor.node.entity_id == self.node.entity_id:
                    anc = '-O  '

        return '%.5s  %10.3g [%6s] %s %s' % (self.node.entity_id, self.magnitude, self.node.direction,
                                             anc, self.name)

    @classmethod
    def from_fragment_flow(cls, ff, group=None, save_unit_scores=False):
        """
        The ff is a FragmentFlow generated during a traversal (or tree crawl)
        :param ff:
        :param group:
        :param save_unit_scores: score_cache must generally not be returned unless some aggregation condition is met
        :return:
        """
        if ff.fragment.reference_entity is None:
            parent = None
        else:
            parent = ff.fragment.reference_entity.external_ref
        scen, a_scen = ff.match_scenarios
        if scen in (1, '1', True):
            scen = 'observed'
        if group is None:
            group = 'StageName'
        node = FragmentRef.from_fragment(ff.fragment)
        anchor = ff.term.to_anchor(save_unit_scores=save_unit_scores)
        ff_m = cls(parent=parent, node=node, name=ff.name,
                   group=ff.fragment.get(group, ''),
                   magnitude=ff.magnitude, scenario=scen, unit=ff.fragment.flow.unit, node_weight=ff.node_weight,
                   flow_conversion=ff.flow_conversion,
                   is_balance_flow=ff.fragment.is_balance,
                   is_cutoff=anchor.is_null,
                   is_conserved=ff.is_conserved,
                   anchor=anchor, anchor_scenario=a_scen)
        if anchor.descend:
            ff_m.subfragments = [FragmentFlow.from_fragment_flow(f, group=group, save_unit_scores=save_unit_scores)
                                 for f in ff.subfragments]
        return ff_m

    def masquerade(self, masq):
        self.node.masquerade(masq)
        if self.anchor:
            self.anchor.masquerade(masq)
        for sf in self.subfragments:
            sf.masquerade(masq)


"""
Foreground serialization

1- foreground serialization has shown to be sufficient to reproduce models
2- thus, formalize the serialization


"""


class MicroCf(ResponseModel):
    ref_quantity: str
    value: Dict[str, float]  # locale, CF


class Compartment(ResponseModel):
    name: str
    parent: Optional[str] = None
    sense: Optional[str] = None
    synonyms: List[str]


class Flowable(ResponseModel):
    name: str
    synonyms: List[str]


class TermManager(ResponseModel):
    Characterizations: Dict[str, Dict[str, Dict[str, MicroCf]]]  # query qty, flowable, compartment
    Compartments: List[Compartment]
    Flowables: List[Flowable]

    @classmethod
    def null(cls):
        return cls(Characterizations=dict(), Compartments=[], Flowables=[])


'''
class LcTermination(ResponseModel):
    """
    these became anchors
    """
    externalId: str
    origin: str
    direction: Optional[str]
    termFlow: Optional[str]
    descend: Optional[str]
    context: Optional[str]
'''


class Observation(ResponseModel):
    fragment: EntityRef
    scenario: Optional[str] = None
    exchange_value: Optional[float] = None
    units: Optional[str] = None
    name: Optional[str] = None
    anchor: Optional[Anchor] = None

    @classmethod
    def from_name(cls, frag):
        return cls(fragment=EntityRef.from_entity(frag), name=frag.external_ref)

    @classmethod
    def dename(cls, frag):
        return cls(fragment=EntityRef.from_entity(frag), name=frag.uuid)

    @classmethod
    def ev(cls, frag, scenario, value, units=None):
        if units is None:
            units = frag.flow.unit
        return cls(fragment=EntityRef.from_entity(frag), scenario=scenario, exchange_value=value, units=units)

    @classmethod
    def from_anchor(cls, frag, scenario, anchor):
        return cls(fragment=EntityRef.from_entity(frag), scenario=scenario, anchor=anchor)

    def masquerade(self, masq):
        self.fragment.masquerade(masq)
        if self.anchor:
            self.anchor.masquerade(masq)


class LcModel(ResponseModel):
    fragments: List[FragmentEntity]

    @classmethod
    def from_reference_fragment(cls, fragment, save_unit_scores=False):
        """
        Replicates the save_fragments method of the LcForeground provider
        for packing prior to transmission over HTTP
        """
        def _recurse_frags(f):
            _r = [f]
            for _x in f.child_flows:  # child flows are already ordered
                _r.extend(_recurse_frags(_x))
            return _r

        fragments = [FragmentEntity.from_entity(k, save_unit_scores=save_unit_scores) for k in _recurse_frags(fragment)]
        return cls(fragments=fragments)

    def serialize(self):
        """
        Replicates the LcFragment.serialize() operation
        for unpacking and storing upon receipt over HTTP
        :return:
        """
        return {
            'fragments': [
                k.serialize() for k in self.fragments
            ]
        }


class ForegroundMetadata(ResponseModel):
    version_major: int
    version_minor: int
    dataSource: str
    release_notes: Optional[str] = None
    description: Optional[str] = 'none'
    author: str = 'nobody'


class ForegroundRelease(ResponseModel):
    major: bool = False
    notes: Optional[str] = None
    description: Optional[str] = None
    author: Optional[str] = None


class LcForeground(ResponseModel):

    catalogNames: Dict[str, List[str]]  #
    dataSource: str  #
    dataSourceType: str  #
    initArgs: Dict  #
    quantities: List[Dict]  #
    termManager: Optional[TermManager]  #
    models: List[LcModel]
    resources: List[ResourceSpec]

    @classmethod
    def from_foreground_archive(cls, ar, save_unit_scores=False):
        """
        A simple function to construct a serialized foreground for transmission to an oryx server
        :param ar: an LcForeground archive
        :param save_unit_scores: subject to permission
        :return: an LcForeground model
        """
        j = ar.serialize(characterizations=True, values=True, domesticate=True)
        ms = [LcModel.from_reference_fragment(f, save_unit_scores=save_unit_scores) for f in ar.fragments()]
        rs = []
        return cls(resources=rs, models=ms, **j)


class ForegroundLciaResult(LciaResult):
    def masquerade(self, masq):
        if self.quantity.origin in masq:
            self.quantity.origin = masq[self.quantity.origin]
        for c in self.components:
            if c.origin in masq:
                c.origin = masq[c.origin]
        for s in self.summaries:
            if s.origin in masq:
                s.origin = masq[s.origin]



import unittest
import os
from shutil import rmtree
from uuid import uuid4

from ...entities.fragment_editor import create_fragment
from .. import LcForeground
from antelope import CatalogRef  # , EntityRefMergeError  ## merge error no longer!

WORKING_DIR = os.path.join(os.path.dirname(__file__), 'test-foreground')

test_ref = 'test.foreground'
flow_uuid = str(uuid4())
frag_uuid = str(uuid4())
frag_ext_name = 'the bodacious reference fragment'

a_different_frag_uuid = str(uuid4())
a_different_frag_ref = 'a completely separate reference fragment'

mass_json = {
    'externalId': 'kg',
    'referenceUnit': 'kg',
    'entityType': 'quantity'
}

flow_json = {
    'externalId': flow_uuid,
    'entityType': 'flow',
    'origin': test_ref,
    'CasNumber': '',
    'referenceQuantity': 'kg'
}

frag_json = [{
    'entityId': a_different_frag_uuid,
    'externalId': a_different_frag_ref,
    'entityType': 'fragment',
    'origin': test_ref,
    'flow': flow_uuid,
    'direction': 'Input',
    'parent': None,
    'isPrivate': False,
    'isBackground': False,
    'isBalanceFlow': False,
    'exchangeValues': {
        '0': 1.0,
        '1': 1.0
    },
    'tags': {
        'Comment': 'this is the fragment the test has made'
    },
    'terminations': {
        'default': {}
    }
}]


"""
Foreground unit testing. What does the foreground do?

on top of a basic archive, it:
 - stores, serializes, deserializes fragments
 - allows to name fragments
 - lists fragments
 - requires all entities to either be local or catalog refs
 * some specialty things like finding terminations nd deleting fragments that are not used
 
For now, the only thing we want to test is the renaming of fragments-- after which either uuid or name should retrieve
the fragment.

Then, this will change somewhat when we upgrade fragments to use links instead of uuids, to allow for different 
instances of the same uuid to be stored in the same foreground.
"""


class LcForegroundTestCase(unittest.TestCase):
    """
    Bucking convention to make this a sequential test, because the operations necessarily depend on one another
    """
    @classmethod
    def setUpClass(cls):
        rmtree(WORKING_DIR, ignore_errors=True)
        cls.fg = LcForeground(WORKING_DIR, ref=test_ref)
        cls.fg.entity_from_json(mass_json)
        cls.fg.entity_from_json(flow_json)

    def test_0_retrieve_flow_by_uuid(self):
        myflow = self.fg[flow_uuid]
        self.assertIsNotNone(myflow)

    def test_1_make_fragment(self):
        myflow = self.fg[flow_uuid]
        frag = create_fragment(myflow, 'Output', uuid=frag_uuid, comment='Test Fragment')
        self.fg.add(frag)

    def test_2_retrieve_fragment(self):
        frag = self.fg[frag_uuid]
        self.assertIsNotNone(frag)
        self.assertEqual(frag['Comment'], 'Test Fragment')

    def test_3_name_fragment(self):
        frag = self.fg[frag_uuid]
        self.assertIsNotNone(frag)
        self.fg.name_fragment(frag, frag_ext_name)
        self.assertIs(frag, self.fg[frag_ext_name])

    def test_4_uuid_still_works(self):
        self.assertIs(self.fg[frag_uuid], self.fg[frag_ext_name])

    def test_5_deserialize_named_fragment(self):
        self.fg._do_load(frag_json)
        self.assertEqual(self.fg[a_different_frag_ref].uuid, a_different_frag_uuid)

    def test_6_save_foreground(self):
        self.fg.save()
        new_fg = LcForeground(WORKING_DIR)
        new_fg.make_interface('basic')  # must now trigger fragment load- this happens automatically in query apparatus
        self.assertEqual(self.fg[flow_uuid], new_fg[flow_uuid])
        self.assertEqual(self.fg[a_different_frag_ref].uuid, new_fg[a_different_frag_uuid].uuid)

    def test_7_different_origins(self):
        my_id = uuid4()  # BaseRef SHOULD cast this to str in external_ref assignment, but as of 0.2.1 it doesn't
        f_ref_1 = CatalogRef('fictitious.origin.v1', my_id, entity_type='flow')
        f_ref_2 = CatalogRef('fictitious.origin.v2', my_id, entity_type='flow')
        self.fg.add(f_ref_1)
        '''
        with self.assertRaises(EntityRefMergeError):
            # this will give an error that CatalogRefs can't merge
            self.fg.add(q_ref_2)
        '''
        self.fg.add(f_ref_2)
        self.assertIs(self.fg[f_ref_2.link], f_ref_2)
        self.assertIs(self.fg[f_ref_1.external_ref], f_ref_2)
        self.assertIs(self.fg[f_ref_1.link], f_ref_1)

    @classmethod
    def tearDownClass(cls):
        rmtree(WORKING_DIR)


if __name__ == '__main__':
    unittest.main()



import json
from collections import defaultdict
from time import time
from antelope import CatalogRef, BasicQuery, comp_dir, NoReference, NoAccessToEntity

from antelope_core.archives import BasicArchive, LC_ENTITY_TYPES
# from lcatools.fragment_flows import FragmentFlow

from ..lc_foreground import FOREGROUND_ENTITY_TYPES
from ...interfaces.iforeground import AntelopeForegroundInterface
from ...refs.fragment_ref import FragmentRef
from ...fragment_flows import FragmentFlow, GhostFragment

from ...terminations import FlowTermination, MissingFlow
from antelope_core.characterizations import DuplicateCharacterizationError
from math import isclose

from .foreground import AntelopeV1ForegroundImplementation
from .quantity import AntelopeQuantityImplementation
from .index import AntelopeIndexImplementation
from .exceptions import AntelopeV1Error

import requests


ANTELOPE_ENTITY_TYPES = LC_ENTITY_TYPES + tuple(filter(lambda k: k not in LC_ENTITY_TYPES, FOREGROUND_ENTITY_TYPES))


try:
    # from urllib.request import urlopen, urljoin
    from urllib.parse import urlparse, urljoin
except ImportError:
    # from urllib2 import urlopen
    from urlparse import urlparse, urljoin


def remote_ref(url):
    p = urlparse(url)
    ref = p.netloc
    for k in p.path.split('/'):
        if len(k) > 0:
            ref = '.'.join([ref, k])
    return ref


class GhostFragmentv1(GhostFragment):
    def __init__(self, parent, top, flow, direction):
        self._top = top
        super(GhostFragmentv1, self).__init__(parent, flow, direction)

    def top(self):
        return self._top


class AntelopeV1Query(BasicQuery, AntelopeForegroundInterface):
    def get_reference(self, external_ref):
        try:
            return super(AntelopeV1Query, self).get_reference(external_ref)
        except NoAccessToEntity:
            raise NoReference(self.origin, external_ref)


class DeferredProcessComment(object):
    def __init__(self, request, process_id):
        super(DeferredProcessComment, self).__init__()
        self._request = request
        self._ppid = process_id
        self._comment = None

    @property
    def comment(self):
        if self._comment is None:
            self._comment = self._request(self._ppid)
        return self._comment

    @property
    def cached(self):
        return self._comment is not None

    def __str__(self):
        return self.comment


class AntelopeV1Client(BasicArchive):
    """
    Provider class for .NET-era Antelope servers.  The basic function is to rely on the interface whenever possible
    but also to cache inventory and LCIA results locally as catalog refs.
    """
    _entity_types = ANTELOPE_ENTITY_TYPES

    def __init__(self, source, ref=None, **kwargs):
        """

        :param source:
        :param catalog:
        :param kwargs:
        """
        if ref is None:
            ref = remote_ref(source)

        self._set_query()

        super(AntelopeV1Client, self).__init__(source, ref=ref, **kwargs)

        self._s = requests.Session()

        # a set of dicts where the key is a string-formatted integer and the value is an entity_ref
        self._endpoints = dict()
        self._cached = defaultdict(dict)

        self._fetched_all = {
            'process': False,
            'flow': False,
            'flowproperty': False,
            'lciamethod': False,
            'fragment': False
        }
        # print('Fetching all flow properties')  ## this makes the difference between a 20 sec test and a 55-sec test
        for fp in self.get_endpoint('flowproperties', cache=False):
            self._parse_and_save_entity(fp)
        # print('done') ## super slow DiscountASP gotta get away
        self._fetched_all['flowproperty'] = True

    def _set_query(self):
        self._query = AntelopeV1Query(self)

    def make_interface(self, iface):
        if iface == 'index':
            return AntelopeIndexImplementation(self)
        elif iface == 'quantity':
            return AntelopeQuantityImplementation(self)
        elif iface == 'foreground':
            return AntelopeV1ForegroundImplementation(self)
        else:
            return super(AntelopeV1Client, self).make_interface(iface)

    def _make_ref(self, external_ref, entity_type, **kwargs):
        if entity_type == 'fragment':
            return FragmentRef(external_ref, self.query, **kwargs)
        r = CatalogRef.from_query(external_ref, self.query, entity_type, **kwargs)
        return r

    def entities_by_type(self, entity_type):
        if entity_type == 'process':
            endp = 'processes'
        elif entity_type == 'flow':
            endp = 'flows'
        elif entity_type == 'flowproperty':
            endp = 'flowproperties'
        elif entity_type == 'lciamethod':
            endp = 'lciamethods'
        elif entity_type == 'fragment':
            endp = 'fragments'
        else:
            raise ValueError('Unknown entity type %s' % entity_type)

        if self._fetched_all[entity_type]:
            for k, v in sorted(self._cached[endp].items()):
                if entity_type == 'flowproperty':
                    if v.is_lcia_method:
                        continue
                elif entity_type == 'lciamethod':
                    if not v.is_lcia_method:
                        continue
                yield v
        else:
            for j in self.get_endpoint(endp, cache=False):
                yield self._parse_and_save_entity(j)
            self._fetched_all[entity_type] = True

    def fragments(self, **kwargs):
        return self.entities_by_type('fragment')

    def get_uuid(self, ext_ref):
        ent = self.retrieve_or_fetch_entity(ext_ref)
        return ent.uuid

    def _load_all(self, **kwargs):
        raise AntelopeV1Error('Cannot load all entities from remote Antelope server')

    def get_endpoint(self, endpoint, cache=True):
        """
        Generally we do not want to cache responses that get parsed_and_saved because that process pop()s
        all the relevant information out of them.
        :param endpoint:
        :param cache:
        :return:
        """
        if endpoint in self._endpoints:
            return self._endpoints[endpoint]

        self._print('Fetching %s from remote server' % endpoint)
        url = urljoin(self.source, endpoint)
        print('loading URL %s' % url)
        t = time()
        j = json.loads(self._s.get(url).content)
        el = time() - t
        print('done [%.2f sec]' % el)

        if cache:
            self._endpoints[endpoint] = j
        return j

    def get_stage_name(self, stage_id):
        stgs = self.get_endpoint('stages')
        stage_id = int(stage_id)
        if stgs[stage_id - 1]['fragmentStageID'] == stage_id:
            return stgs[stage_id - 1]['name']
        try:
            return next(j['name'] for j in stgs if j['fragmentStageID'] == stage_id)
        except StopIteration:
            raise ValueError('Unknown fragment stage ID %d' % stage_id)

    def _ref_to_key(self, key):
        return key

    def _fetch(self, entity, **kwargs):
        j = self.get_endpoint(entity, cache=False)[0]
        return self._parse_and_save_entity(j)

    def _get_comment(self, processId):
        target = '/'.join(['processes', processId, 'comment'])
        j = self.get_endpoint(target)
        return j['comment']

    def _get_impact_category(self, cat_id):
        cats = self.get_endpoint('impactcategories')
        cat_id = int(cat_id)
        try:
            return next(j['name'] for j in cats if j['impactCategoryID'] == cat_id)
        except StopIteration:
            raise ValueError('Unknown impact category ID %d' % cat_id)

    def fetch_flows(self, fragment):
        for f in self.get_endpoint('%s/flows' % fragment, cache=False):
            ext_ref = 'flows/%d' % f['flowID']
            if ext_ref in self._entities:
                continue
            self._parse_and_save_entity(f)

    def make_fragment_flow(self, ff):
        """
            A fragment traversal generates an array of FragmentFlow objects.

    X    "fragmentID": 8, - added by antelope
    X    "fragmentStageID": 80,

    f    "fragmentFlowID": 167,
    f    "name": "UO Local Collection",
    f    "shortName": "Scenario",
    f    "flowID": 371,
    f    "direction": "Output",
    f    "parentFragmentFlowID": 168,
    f    "isBackground": false,

    w    "nodeWeight": 1.0,

    t    "nodeType": "Process",
    t    "processID": 62,

    *    "isConserved": true,
    *    "flowPropertyMagnitudes": [
      {
        "flowPropertyID": 23,
        "unit": "kg",
        "magnitude": 1.0
      }
    ]

        Need to:
         * create a termination
         * create a fragment ref
         * extract node weight
         * extract magnitude
         * extract is_conserved

        :param ff: .from_antelope_v1(ff, self.query)
         JSON-formatted fragmentflow, from a v1 .NET antelope instance.  Must be modified to include StageName
         instead of fragmentStageID
        :return:
        """
        """
        This needs to be in-house because it uses the query mechanism
        :param ff:
        :return:
        """
        fpms = ff['flowPropertyMagnitudes']
        ref_mag = fpms[0]
        magnitude = ref_mag['magnitude']
        flow = self.query.get('flows/%s' % ff['flowID'])
        for fpm in fpms[1:]:
            mag_qty = self.query.get('flowproperties/%s' % fpm['flowPropertyID'])
            if fpm['magnitude'] == 0:
                val = 0
            else:
                val = fpm['magnitude'] / magnitude
            try:
                flow.characterize(mag_qty, value=val)
            except DuplicateCharacterizationError:
                if not isclose(mag_qty.cf(flow), val):
                    raise ValueError('Characterizations do not match: %g vs %g' % (mag_qty.cf(flow), val))

        dirn = ff['direction']

        if 'parentFragmentFlowID' in ff:
            parent_ref = 'fragments/%s/fragmentflows/%s' % (ff['fragmentID'], ff['parentFragmentFlowID'])
            top = self.query.get('fragments/%s' % ff['fragmentID'])
            frag = GhostFragmentv1(parent_ref, top, flow, dirn)  # distinctly not reference

        else:
            frag = self.query.get('fragments/%s' % ff['fragmentID'])

        node_type = ff['nodeType']
        nw = ff['nodeWeight']
        if magnitude == 0:
            inbound_ev = 0
        else:
            inbound_ev = magnitude / nw

        if node_type == 'Process':
            term_node = self.query.get('processes/%s' % ff['processID'])
            term = FlowTermination(frag, term_node, term_flow=flow, inbound_ev=inbound_ev,
                                   _direction=comp_dir(dirn))  # antelope v1 processes do not have reference flows!
        elif node_type == 'Fragment':
            term_node = self.query.get('fragments/%s' % ff['subFragmentID'])
            try:
                term = FlowTermination(frag, term_node, term_flow=flow, inbound_ev=inbound_ev)
            except MissingFlow:
                term = FlowTermination(frag, term_node, term_flow=None, inbound_ev=inbound_ev)
        else:
            term = FlowTermination.null(frag)
        if 'isConserved' in ff:
            conserved = ff['isConserved']
        else:
            conserved = False

        return FragmentFlow(frag, magnitude, nw, term, conserved)

    '''
    Entity handling
    '''
    def retrieve_or_fetch_entity(self, key, **kwargs):
        """
        A key in v1 antelopev1 is 'datatype/index' where index is a sequential number starting from 1.  The archive also
        caches references locally by their uuid, so retrieving an entity by uuid will work if the client has already
        come across it.
        :param key:
        :param kwargs:
        :return:
        """
        ent = None

        if key in self._entities:
            return self._entities[key]

        parts = key.split('/')
        if len(parts) != 2:
            raise KeyError('%s: Key must be of the format "entity_type/id"' % key)
        if parts[1] in self._cached[parts[0]]:
            ent = self._cached[parts[0]][parts[1]]
        if ent is not None:
            return ent
        else:
            return self._fetch(key, **kwargs)

    def _parse_and_save_entity(self, j):
        """
        Take a json object obtained from a query and use it to create an entity ref
        :param j:
        :return:
        """
        func = {'process': self._parse_and_save_process,
                'flow': self._parse_and_save_flow,
                'lciamethod': self._parse_and_save_lcia,
                'flowproperty': self._parse_and_save_fp,
                'fragment': self._parse_and_save_fragment}[j.pop('resourceType').lower()]
        j.pop('links')
        obj = func(j)
        self.add_entity_and_children(obj)
        return obj

    def _parse_and_save_process(self, j):
        process_id = str(j.pop('processID'))
        ext_ref = 'processes/%s' % process_id
        j['Name'] = j.pop('name')
        j['SpatialScope'] = j.pop('geography')
        j['TemporalScope'] = j.pop('referenceYear')
        j['Comment'] = DeferredProcessComment(self._get_comment, process_id)
        ref = self._make_ref(ext_ref, 'process', reference_entity=[], **j)
        self._cached['processes'][process_id] = ref
        return ref

    def _parse_and_save_flow(self, j):
        flow_id = str(j.pop('flowID'))
        ext_ref = 'flows/%s' % flow_id
        j['Name'] = j.pop('name')
        j['CasNumber'] = j.pop('casNumber', '')
        j['Compartment'] = [j.pop('category')]
        ref_qty = self.retrieve_or_fetch_entity('flowproperties/%s' % j.pop('referenceFlowPropertyID'))
        ref = self._make_ref(ext_ref, 'flow', reference_entity=ref_qty, **j)
        self._cached['flows'][flow_id] = ref
        return ref

    def _parse_and_save_lcia(self, j):
        lm_id = str(j.pop('lciaMethodID'))
        ext_ref = 'lciamethods/%s' % lm_id
        j['Name'] = j.pop('name')
        j['Method'] = j.pop('methodology')
        rfp = j.pop('referenceFlowProperty')
        ref_unit = rfp['referenceUnit']
        j.pop('referenceFlowPropertyID')
        j['Indicator'] = rfp['name']
        j['Category'] = self._get_impact_category(j.pop('impactCategoryID'))
        ref = self._make_ref(ext_ref, 'quantity', reference_entity=ref_unit, **j)
        self._cached['lciamethods'][lm_id] = ref
        return ref

    def _parse_and_save_fp(self, j):
        fp_id = str(j.pop('flowPropertyID'))
        ext_ref = 'flowproperties/%s' % fp_id
        j['Name'] = j.pop('name')
        ref_unit = j.pop('referenceUnit')
        ref = self._make_ref(ext_ref, 'quantity', reference_entity=ref_unit, **j)
        self._cached['flowproperties'][fp_id] = ref
        return ref

    def _parse_and_save_fragment(self, j):
        frag_id = str(j.pop('fragmentID'))
        ext_ref = 'fragments/%s' % frag_id
        j['Name'] = j.pop('name')
        dirn = j.pop('direction')
        flow = self.retrieve_or_fetch_entity('flows/%s' % j.pop('termFlowID'))
        ref = self._make_ref(ext_ref, 'fragment', flow=flow, direction=dirn, **j)
        self._cached['fragments'][frag_id] = ref
        return ref



from antelope import comp_dir, ExchangeRef

from .terminations import FlowTermination, UnCachedScore, UnresolvedAnchor
from .models import DescendSpec
from antelope_core.lcia_results import LciaResult, DetailedLciaResult, SummaryLciaResult

from collections import defaultdict
import uuid
import logging


class FragmentInventoryDeprecated(Exception):
    """
    The "inventory" term for fragmnts is deprecated.  "unit_inventory" is now "unit_flows" and "inventory"
    is what it has always been: "cutoffs".  And the old "cutoffs" is gone because it literally never worked.
    """
    pass


class CumulatingFlows(Exception):
    """
    when a fragment includes multiple instances of the reference flow having consistent (i.e. not complementary)
    directions. Not handled in subfragment traversal bc no valid test case
    """
    pass


class FragmentFlow(object):
    """
    A FragmentFlow is a an immutable record of a traversal query. essentially an enhanced NodeCache record which
    can be easily serialized to an antelope fragmentflow record.

    A fragment traversal generates an array of FragmentFlow objects.

    X    "fragmentID": 8, - added by antelope
    X    "fragmentStageID": 80,

    f    "fragmentFlowID": 167,
    f    "name": "UO Local Collection",
    f    "shortName": "Scenario",
    f    "flowID": 371,
    f    "direction": "Output",
    f    "parentFragmentFlowID": 168,
    f    "isBackground": false,

    w    "nodeWeight": 1.0,

    t    "nodeType": "Process",
    t    "processID": 62,

    *    "isConserved": true,
    *    "flowPropertyMagnitudes": [
      {
        "flowPropertyID": 23,
        "unit": "kg",
        "magnitude": 1.0
      }
    ]

    """
    '''
    @classmethod
    def from_antelope_v1(cls, j, query):
        """
        Need to:
         * create a termination
         * create a fragment ref
         * extract node weight
         * extract magnitude
         * extract is_conserved
        :param j: JSON-formatted fragmentflow, from a v1 .NET antelope instance.  Must be modified to include StageName
         instead of fragmentStageID
        :param query: an antelope v1 catalog query
        :return:
        """
        return cls(frag, magnitude, nw, term, conserved)

    @classmethod
    def ref_flow(cls, parent, use_ev):
        """

        :param parent:
        :param use_ev: required to create reference flows from fragment refs
        :return:
        """
        fragment = GhostFragment(parent, parent.flow, comp_dir(parent.direction))
        term = FlowTermination.null(fragment)
        return cls(fragment, use_ev, 1.0, term,
                   parent.is_conserved_parent)
    '''

    @classmethod
    def cutoff(cls, parent, flow, direction, magnitude, is_conserved=False):
        fragment = GhostFragment(parent, flow, direction)
        term = FlowTermination.null(fragment)
        return cls(fragment, magnitude, magnitude, term, is_conserved)

    def __init__(self, fragment, magnitude, node_weight, term, is_conserved, match_ev=None, match_term=None,
                 flow_conversion=1.0):
        """

        :param fragment:
        :param magnitude:
        :param node_weight: flow (or balance) magnitude * flow conversion / anchor inflow magnitude
        :param term:
        :param is_conserved:
        :param match_ev:
        :param match_term:
        :param flow_conversion: [1.0] stored in the FragmentFlow for information purposes only. Negative value
         indicates a direction change at the anchor (driven anchor)
        """
        # TODO: figure out how to cache + propagate scenario applications through aggregation ops
        self.fragment = fragment
        self.magnitude = magnitude
        self.node_weight = node_weight
        self.flow_conversion = flow_conversion
        self.term = term
        self.is_conserved = is_conserved
        self._subfrags_params = ()
        self.match_scenarios = (match_ev, match_term)
        self._superfrag = None
        self.uuid = str(uuid.uuid4())

    @property
    def superfragment(self):
        return self._superfrag

    @property
    def subfragments(self):
        if self.term.is_subfrag:  # and (self.term.descend is False):
            try:
                return self._subfrags_params[0]
            except IndexError:
                return []
        return []

    @property
    def subfragment_scenarios(self):
        try:
            return self._subfrags_params[1]
        except IndexError:
            return None

    def aggregate_subfragments(self, subfrags, scenarios=None):
        """
        We need to save the full traversal specification (
        :param subfrags:
        :param scenarios:
        :return:
        """
        for f in subfrags:
            f._superfrag = self
        self._subfrags_params = (subfrags, scenarios)

    def scale(self, x):
        self.node_weight *= x
        self.magnitude *= x

    @property
    def name(self):
        return self.term.name

    def screen_name(self, length=80):
        """
        auto-compact for display

        :return:
        """
        name = self.name
        if len(name) > length:
            name = name[:(length - 18)] + '....' + name[-14:]
        return name

    def __str__(self):
        return '%.5s  %10.3g [%6s] %s %s' % (self.fragment.uuid, self.magnitude, self.fragment.direction,
                                             self.term, self.name)

    def __add__(self, other):
        if isinstance(other, FragmentFlow):
            if other.fragment.uuid != self.fragment.uuid:
                raise ValueError('Fragment flows do not belong to the same fragment')
            mag = other.magnitude
            nw = other.node_weight
            if not self.term == other.term:
                raise ValueError('These fragment flows are differently terminated')

            if mag * self.node_weight != (self.magnitude * nw):  # formally if m*N/M*n != 1.0:
                raise ValueError('These fragment flows cannot be combined because their implicit evs do not match')
            conserved = self.is_conserved and other.is_conserved

            mod_mag = mag * other.flow_conversion / self.flow_conversion
            """
        elif isinstance(other, DetailedLciaResult):
            print('DEPRECATED: adding FragmentFlow to DetailedLciaResult')
            if other.exchange.process is not self.fragment:
                raise ValueError('FragmentFlow and DetailedLciaResult do not belong to the same fragment')
            nw = other.exchange.value
            mag = nw
            conserved = False
        elif isinstance(other, SummaryLciaResult):
            print('DEPRECATED: adding FragmentFlow to SummaryLciaResult')
            if other.entity is not self.fragment:
                raise ValueError('FragmentFlow and SummaryLciaResult do not belong to the same fragment')
            nw = other.node_weight
            mag = nw
            conserved = False
            """
        else:
            raise TypeError("Don't know how to add type %s to FragmentFlow\n %s\n to %s" % (type(other), other, self))
        # don't check unit scores-- ?????
        new = FragmentFlow(self.fragment, self.magnitude + mod_mag, self.node_weight + nw,
                           self.term, conserved, flow_conversion=self.flow_conversion)
        return new

    def get(self, item, default=None):
        return self.fragment.get(item, default=default)

    def __getitem__(self, item):
        return self.fragment.__getitem__(item)

    def __eq__(self, other):
        """
        FragmentFlows are equal if they have the same fragment and termination.  Formerly magnitude too but why?
        answer why: because if not, then two traversals from two different scenarios can appear equal
        answer why not: because of LciaResult issues, presumably- let's put this on the list for later
        :param other:
        :return:
        """
        if not isinstance(other, FragmentFlow):
            return False
        return self.fragment == other.fragment and self.term == other.term  # and self.node_weight == other.node_weight

    def __hash__(self):
        return hash(self.fragment)

    @property
    def ref_unit(self):
        return self.fragment.flow.unit


def group_ios(parent, ffs, include_ref_flow=True, passthru_threshold=0.45):
    """
    Utility function for dealing with a traversal result (list of FragmentFlows)
    Creates a list of cutoff flows from the inputs and outputs from a fragment traversal.

    Returns two lists of FragmentFlows: external (inputs+outputs) and internal to the traversal.

    Pass-Thru Flows and Autoconsumption

    There is a challenge in dealing a fragment whose inventory contains the fragment's own reference flow.  The two
    broad interpretations of this are that the substance is "passing through" the fragment, or that the fragment
    induces further consumption of its reference flow or demand for its own service.  Both approaches are logical in
    different circumstances, and it is difficult to tell the difference in certain cases.

    There are eight cases: whether the flow directions are complementary or cumulative, whether the flow grows or
    shrinks, whether the reference direction is input or output.  Pictorially, with <== and <-- indicating larger and
    smaller flows respectively; * representing the reference fragment; --: representing the child flows

    A.1.a  <== *--: <--   augmentation?             B.1.  ==> *--: <--  cumulating sink; NONSENSICAL
        b  <-= *--  <--   autoconsumption?

    A.2.   <-- *--: <==   depletion / yield loss    B.2.  --> *--: <==  cumulating sink; NONSENSICAL
                          induced load NONSENSICAL

    A.3.a  ==> *--: -->   depletion / yield loss?   B.3.  <== *--: -->  cumulating source; NONSENSICAL
        b  -=> *--  -->   induced self-load? (probably nonsensical)

    A.4.   --> *--: ==>   augmentation              B.4.  <-- *--: ==>  cumulating source; NONSENSICAL
                          autoconsumption NONSENSICAL

    In all the (B) cases where the two flows are oppositely directed, the driven fragment multiplies its own effect, and
    these fragment designs are considered nonsensical and are not allowed.

    In cases A.2 and A.4, autoconsumption is nonsensical because it would require the induced amount to be modeled as
    the reference flow, but the induced amount cannot be induced by itself.  So these are automatically interpreted as
    pass-through (i.e. A.2 is depletion of a passed-through flow; A.4 is accumulation of a passed-through flow).

    In cases A.1 and A.3 however, either could apply:

    A.1.a A component is supplied after being further assembled
    A.1.b grid power / pipeline service consumes its own output in delivering its service

    A.3.a A recycling stream is purified of contaminants
    A.3.b A treatment process generates some of its own waste for treatment (??) maybe nonsensical but we allow it

    In these cases, helpless to divine the modeler's intent, we apply a threshold.  If the induced amount is below this
    threshold as a fraction of the reference flow, it will be taken as an autoconsumption process-- the induced flow
    will be subsumed into the reference flow, reducing its magnitude.

    If the induced amount exceeds this threshold, the process will be taken to be an augmentation / depletion pass-
    through process.  The threshold is set to 0.33, just because it seems safe to assume that an induced load will not
    exceed 33% of the direct load, or that a process will not augment or deplete its own flow by 3x.

    As a reminder, if you need to model such a process, it is easy to do- just give the input and output distinct flows!

    :param parent: the node generating the cutoffs
    :param ffs: a list of fragment flows resulting from a traversal of the parent
    :param include_ref_flow: [True] whether to include the reference fragment and adjust for autoconsumption
    :param passthru_threshold: [0.33] smaller than this is treated as autoconsumption / induced load
    :return: [list of grouped IO flows], [list of internal non-null flows]
    """
    out = defaultdict(float)  # accumulates net total
    pos_mag = defaultdict(float)  # accumulates +ve
    neg_mag = defaultdict(float) # accumulates -ve
    dirs = dict()
    # cons = dict()
    internal = []
    external = []
    for ff in ffs:
        if ff.term.is_null:
            # accumulate IO flows according to the first seen direction
            if ff.fragment.flow not in dirs:
                dirs[ff.fragment.flow] = ff.fragment.direction
            # and correct the signs of subsequent flows
            if ff.fragment.direction == dirs[ff.fragment.flow]:
                magnitude = ff.magnitude
            else:
                magnitude = -ff.magnitude

            ''' # ditto conservation status  # but this doesn't work
            if ff.fragment.flow not in cons:
                cons[ff.fragment.flow] = ff.is_conserved
            else:
                if ff.is_conserved ^ cons[ff.fragment.flow]:
                    print('** Surprise! conservation status not shared among matching flows %s' % ff.fragment.uuid)
            '''

            out[ff.fragment.flow] += magnitude
            if magnitude > 0:  # separate test from above because we are ignoring relative-directions
                pos_mag[ff.fragment.flow] += magnitude  # positive
            else:
                neg_mag[ff.fragment.flow] -= magnitude  # positive
        else:
            internal.append(ff)

    # now deal with reference flow-- trivial fragment should wind up with two equal-and-opposite [pass-through] flows
    ref_frag = parent.top()
    if include_ref_flow:
        # ref_cons = ref_frag.is_conserved_parent
        ref_mag = ffs[0].magnitude
        if ref_frag.flow in out:  # either pass through or autoconsumption
            ref_frag.dbg_print('either pass through or autoconsumption')
            val = out[ref_frag.flow]
            auto_dirn = dirs[ref_frag.flow]
            if val < 0:
                auto_dirn = comp_dir(auto_dirn)  # this is rare, but
            # else:
            #    auto_dirn = 'Input'
            """
            If the directions are cumulating, we raise an error- bad fragment design. 
            
            If directions are complementary [meaning equal since ref flow dirn is w.r.t. parent], this means the 
            reference outflow is an inventory inflow, or the reference inflow is an inventory outflow.
            
            We apply a threshold, where if the induced flow is below the threshold w.r.t. the ref flow, autoconsumption
            is applied.
            """
            if auto_dirn == ref_frag.direction:  # equality here means complementary flows - direction is w.r.t. parent
                # case A here
                if abs(val) < (passthru_threshold * ref_mag):
                    # A.1.b and A.3.b
                    ref_frag.dbg_print('autoconsumption %g %g' % (val, ref_mag))
                    # autoconsumption, the inventory flow is subsumed by the ref_flow; direction sense should reverse
                    if auto_dirn == 'Output':
                        out[ref_frag.flow] += ref_mag
                    else:
                        out[ref_frag.flow] -= ref_mag
                    if ref_mag > 0:
                        pos_mag[ref_frag.flow] += ref_mag
                    else:
                        neg_mag[ref_frag.flow] -= ref_mag

                else:
                    # A.1.a, A.3.a, A.2, A.4
                    ref_frag.dbg_print('pass thru no effect %g %g' % (val, ref_mag))
                    # pass-thru: pre-initialize external with the reference flow, having the opposite direction
                    external.append(FragmentFlow.cutoff(parent, ref_frag.flow, comp_dir(auto_dirn), ref_mag)) #,
                                                        #is_conserved=ref_cons))
            else:
                # all case B
                ref_frag.dbg_print('cumulation! %g %g' % (val, ref_mag))
                # cumulation: the directions are both the same... should they be accumulated?  not handled
                raise CumulatingFlows('%s' % parent)
                # external.append(FragmentFlow.cutoff(parent, ref_frag.flow, auto_dirn, ref_mag))
        else:
            ref_frag.dbg_print('uncomplicated ref flow')
            # no autoconsumption or pass-through, but we still want the ref flow to show up in the inventory
            external.append(FragmentFlow.cutoff(parent, ref_frag.flow, comp_dir(ref_frag.direction), ref_mag)) #,
                                                #is_conserved=ref_cons))

    for flow, value in out.items():
        direction = dirs[flow]
        abs_mag = max([pos_mag[flow], neg_mag[flow]])
        if value != 0:
            if abs_mag == 0:  # we'd hate to get a ZeroDivisionError-- but this should be impossible
                print('group_ios weird zero abs_mag %.5s (%s: %g, %g)' % (ref_frag.uuid, flow, value, abs_mag))
            elif abs(value) / abs_mag < 1e-12:
                # these have never been controversial-- stop cluttering the console
                # print('Quashing group_ios flow < 1e-12 magnitude (%s: %g, %g)' % (flow, value, abs_mag))
                ref_frag.dbg_print('Quashing balance flow < 1e-12 magnitude (%s: %g, %g)' % (flow, value, abs_mag))
                value = 0.0
        if value < 0:
            direction = comp_dir(direction)
        external.append(FragmentFlow.cutoff(parent, flow, direction, abs(value)))  # , is_conserved=cons[flow]))

    return external, internal


def ios_exchanges(ios, ref=None, scale=1.0):
    if ref is None:
        ref = ios[0].fragment
    frag_exchs = []
    for f in ios:
        if f.magnitude == 0:
            continue
        is_ref = (f.fragment.flow == ref.flow and f.fragment.direction == comp_dir(ref.direction))

        xv = ExchangeRef(ref, f.fragment.flow, f.fragment.direction, value=f.magnitude * scale, is_reference=is_ref)
        frag_exchs.append(xv)
    return sorted(frag_exchs, key=lambda x: (x.direction == 'Input', x.value), reverse=True)


def frag_flow_lcia(fragmentflows, quantity_ref, scenario=None, descend_all=None, descend_spec=None, **kwargs):
    """
    Recursive function to compute LCIA of a traversal record contained in a set of Fragment Flows.
    Note: refresh is no longer supported during traversal
    :param fragmentflows:
    :param quantity_ref:
    :param scenario: necessary if any remote traversals are required
    :param descend_all: if True or False, disregard frag-specific descend settings in favor of spec
    :param descend_spec: a DescendSpec object (overrides above)
    :return:
    """
    if descend_spec is None:
        descend_spec = DescendSpec(descend_all=descend_all)

    result = LciaResult(quantity_ref, scenario=str(scenario))
    # _first_ff = True  # I have no idea what problem this was meant to solve
    for ff in fragmentflows:
        # _recursive_remote = False
        if ff.term.is_null:
            continue

        node_weight = ff.node_weight
        if node_weight == 0:
            continue

        if ff.term.direction == ff.fragment.direction:
            # if the directions collide (rather than complement), the term is getting run in reverse
            node_weight *= -1

        # if we have subfragments, use them
        if len(ff.subfragments) == 0:  # always true for: contexts, processes, remote subfragments
            try:
                v = ff.term.score_cache(quantity=quantity_ref, **kwargs)

                # if we reach here, then we have successfully retrieved a cached unit score and we are done
                if not v.is_null:
                    result.add_summary(ff.uuid, ff, node_weight, v)

                # _first_ff = False
                continue

            except UnresolvedAnchor:
                result.add_missing(ff.uuid, ff.term.term_node, node_weight)
                # _first_ff = False
                continue

            except UnCachedScore:
                # a subfragment with no stored subfragments and no cached score: we gotta ask
                v = ff.term.term_node.fragment_lcia(quantity_ref, scenario=scenario)
                # _recursive_remote = True  # skip this noise

        else:
            v = frag_flow_lcia(ff.subfragments, quantity_ref, scenario=ff.subfragment_scenarios,
                               descend_spec=descend_spec, **kwargs)
            if v.is_null:
                continue

        # if we arrive here, we have a unit score from a subfragment

        if descend_spec.descend_ff(ff):
            if v.has_summaries:
                for k in v.keys():
                    c = v[k]
                    result.add_summary(k, c.entity, c.node_weight * node_weight, c.internal_result)
            else:
                result.add_summary(ff.uuid, ff, node_weight, v)
        else:
            result.add_summary(ff.uuid, ff, node_weight, v)

        # if _first_ff and _recursive_remote:
        #     if len(fragmentflows) > 1:
        #         logging.warning('Bailing out early despite %d un-handled fragment flows' % (len(fragmentflows)-1))
        #     return result  # bail out
        # _first_ff = False
    return result


class GhostFragment(object):
    """
    A GhostFragment is a non-actual fragment used for reporting and aggregating fragment inputs and outputs
      during traversal.
      We're doing some digging into this because the prior iteration created totally non-functional FragmentFlow models
    """
    def __init__(self, parent, flow, direction):
        self._parent = parent
        self.flow = flow
        self.direction = direction
        self.uuid = str(uuid.uuid4())

    @property
    def origin(self):
        return self.flow.origin

    @property
    def external_ref(self):
        return self.uuid

    @property
    def parent(self):
        return self._parent

    @property
    def reference_entity(self):
        return self._parent

    @property
    def is_reference(self):
        return self._parent is None

    @property
    def is_background(self):
        return False

    @property
    def is_balance(self):
        """
        no way to track this through aggregation (in the present architecture)
        :return:
        """
        return False

    @property
    def entity_type(self):
        return 'fragment'

    @property
    def dirn(self):
        return {
            'Input': '-<-',
            'Output': '=>='
        }[self.direction]

    def top(self):
        return self._parent.top()

    def __str__(self):
        re = self.reference_entity.uuid[:7]
        return '(%s) %s %.5s %s --:   [%s] %s' % (re, self.dirn, self.uuid, self.dirn,
                                                  self.flow.unit, self.flow['Name'])

    def __eq__(self, other):
        """
        the LcEntity equality test is no good for GhostFragments because they replicate the same parent flow.
        We should test if parent, flow, and direction are the same.
        :return:
        """
        if other is None:
            return False
        # if not isinstance(other, LcEntity):  # taking this out so that CatalogRefs and entities can be compared
        #     return False
        try:
            is_eq = (self.reference_entity == other.reference_entity
                     and self.flow == other.flow
                     and self.direction == other.direction)
        except AttributeError:
            is_eq = False
        return is_eq

    def get(self, item, default=None):
        try:
            return self.__getitem__(item)
        except KeyError:
            return default

    def __getitem__(self, item):
        if item == 'parent':
            if self._parent:
                return self._parent.external_ref
            else:
                return None
        try:
            return self.flow[item]
        except KeyError:
            if self._parent:
                return self._parent[item]
            raise



from itertools import chain
import logging

from antelope import EntityNotFound, comp_dir, BackgroundRequired
from ..interfaces.iforeground import AntelopeForegroundInterface  # , ForegroundRequired
from antelope_core.implementations import BasicImplementation
from antelope_core.implementations.quantity import UnknownRefQuantity

from antelope_core.entities.xlsx_editor import XlsxArchiveUpdater
from antelope_core.contexts import NullContext
from antelope_core.entities.quantities import new_quantity
from antelope_core.entities.flows import new_flow

from ..entities.fragments import LcFragment, InvalidParentChild, FragmentBranch
from ..entities.fragment_editor import create_fragment, clone_fragment, _fork_fragment, interpose
from ..models import ForegroundRelease, Anchor


class NotForeground(Exception):
    pass


class UnknownFlow(Exception):
    pass


class AntelopeBasicImplementation(BasicImplementation):
    def get_reference(self, key):
        entity = self._dereference_entity(key)
        if entity.entity_type == 'fragment':
            return [entity.reference()]
        return super(AntelopeBasicImplementation, self).get_reference(key)


class AntelopeForegroundImplementation(BasicImplementation, AntelopeForegroundInterface):
    """
    A foreground manager allows a user to build foregrounds.  This should work with catalog references, rather
    than actual entities.

    This interface should cover all operations for web tool

    To build a fragment, we need:
     * child flows lambda
     * uuid

     * parent or None

     * Flow and Direction (relative to parent)
     * cached exchange value (or 1.0)

     * external_ref is uuid always? unless set-- ForegroundArchive gets a mapping from external_ref to uuid

     * other properties: Name, StageName

    To terminate a fragment we need a catalog ref: origin + ref for:
     * to foreground -> terminate to self
     * to subfragment -> just give uuid
     * background terminations are terminations to background fragments, as at present
     * process from archive, to become sub-fragment
     * flow, to become fg emission

    Fragment characteristics:
     * background flag
     * balance flag

    Scenario variants:
     - exchange value
     - termination
    """
    # _count = 0
    # _frags_with_flow = defaultdict(set)  # we actually want this to be shared among
    # _recursion_check = None

    ''' # NOT YET
    def __getitem__(self, item):
        """
        Let's have some standard etiquette here, people
        :param item:
        :return:
        """
        return self.get_local(item)
    '''

    '''
    Add some useful functions from other interfaces to the foreground
    '''
    @property
    def delayed(self):
        return self._archive.delayed

    @property
    def unresolved(self):
        return self._archive.unresolved

    def catalog_ref(self, *args, **kwargs):
        return self._archive.catalog_ref(*args, **kwargs)
    
    def catalog_query(self, origin, **kwargs):
        return self._archive.catalog_query(origin, **kwargs)

    def apply_xlsx(self, xlsx, quiet=True):
        """
        This is kind of inside baseball as long as xls_tools is not public
        :param self: a resource
        :param xlsx: an Xlrd-like spreadsheet: __getitem__(sheetname); sheet.row(i), sheet.column(j), sheet.nrows, sheet.ncols
        :param quiet:
        :return: nothing to return- user already has the resource
        """
        with XlsxArchiveUpdater(self._archive, xlsx, quiet=quiet, merge='overwrite') as x:
            x.apply()

    def count(self, entity_type):
        return self._archive.count_by_type(entity_type)

    def flows(self, **kwargs):
        for f in self._archive.search('flow', **kwargs):
            yield f

    def targets(self, flow, direction, **kwargs):
        return self.fragments_with_flow(flow, direction, **kwargs)

    '''
    for internal / convenience use
    '''
    def get_canonical(self, quantity):
        """
        By convention, a foreground archive's Term Manager is the catalog's LCIA engine, which is the Qdb of record
        for the foreground.
        :param quantity:
        :return:
        """
        return self._archive.tm.get_canonical(quantity)

    def context(self, item):
        return self._archive.tm[item]

    '''
    def get_context(self, item):
        return self._archive.tm[item]
    '''

    def flowable(self, item):
        return self._archive.tm.get_flowable(item)

    ''' # outmoded by find_term?
    def _grounded_ref(self, ref, check_etype=None):
        """
        Accept either a string, an unresolved catalog ref, a resolved catalog ref, or an entity
        Return an entity or grounded ref
        :param ref:
        :param check_etype:
        :return:
        """
        if hasattr(ref, 'entity_type'):
            if (not ref.is_entity) and (not ref.resolved):  # unresolved catalog ref
                try:
                    ent = self.get(ref.external_ref)
                except EntityNotFound:
                    ent = self._archive.catalog_ref(ref.origin, ref.external_ref)
            else:  # entity or resolved catalog ref
                ent = ref
        else:  # stringable
            try:
                ent = self.get(ref)
            except EntityNotFound:
                ent = self.get_local(ref)
        if check_etype is not None and ent.entity_type != check_etype:
            raise TypeError('%s: Not a %s' % (ref, check_etype))
        return ent
    '''

    '''
    fg implementation begins here
    '''
    def fragments(self, *args, show_all=False, **kwargs):
        if hasattr(self._archive, 'fragments'):
            # we only want reference fragments by default
            for f in self._archive.fragments(*args, show_all=show_all, **kwargs):
                yield f
        else:
            raise NotForeground('The resource does not contain fragments: %s' % self._archive.ref)

    def frag(self, string, **kwargs):
        """
        :param string:
        :param kwargs: many=False
        :return:
        """
        return self._archive.frag(string, **kwargs)

    def frags(self, string, knobs=True):
        """
        Show fragments whose names begin with
        :param string:
        :param knobs: [True] by default, only list named non-reference fragments (knobs).  If knobs=False,
        list reference fragments
        :return:
        """
        if knobs:
            for k in self.knobs():
                if k.external_ref.startswith(string):
                    yield k
        else:
            for k in self.fragments():
                if k.external_ref.startswith(string):
                    yield k

    '''
    Create and modify fragments
    '''
    def new_quantity(self, name, ref_unit=None, **kwargs):
        """

        :param name:
        :param ref_unit:
        :param kwargs:
        :return:
        """
        q = new_quantity(name, ref_unit, origin=self.origin, **kwargs)
        self._archive.add(q)
        return q

    def add_entity_and_children(self, *args, **kwargs):
        """
        This is invoked by xlsx_editor - putatively on an archive- when trying to create quantities that are found
        in the quantity db but not in the local archive.
        Then the model_updater replaces the archive with a foreground implementation, but as far as I can tell the
        foreground implementation NEVER had this method. So I don't understand how it was working before.
        :param args:
        :param kwargs:
        :return:
        """
        return self._archive.add_entity_and_children(*args, **kwargs)  # this is a hack

    def add_or_retrieve(self, external_ref, reference, name, group=None, strict=False, **kwargs):
        """
        Gets an entity with the given external_ref if it exists, , and creates it if it doesn't exist from the args.

        Note that the full spec is mandatory so it could be done by tuple.  With strict=False, an entity will be
        returned if it exists.
        :param external_ref:
        :param reference: a string, either a unit or known quantity
        :param name: the object's name
        :param group: used as context for flow
        :param strict: [False] if True it will raise a TypeError (ref) or ValueError (name) if an entity exists but
        does not match the spec.  n.b. I think I should check the reference entity even if strict is False but.. nah
        :return:
        """
        try:
            t = self.get(external_ref)
            if strict:
                if t.entity_type == 'flow':
                    if t.reference_entity != self.get_canonical(reference):
                        raise TypeError("ref quantity (%s) doesn't match supplied (%s)" % (t.reference_entity,
                                                                                           reference))
                elif t.entity_type == 'quantity':
                    if t.unit != reference:
                        raise TypeError("ref unit (%s) doesn't match supplied (%s)" % (t.unit, reference))
                if t['Name'] != name:
                    raise ValueError("Name (%s) doesn't match supplied(%s)" % (t['Name'], name))
            else:
                if t['Name'] != name:
                    t['Name'] = name
            for k, v in kwargs.items():
                if v is not None:
                    t[k] = v
            return t

        except EntityNotFound:
            try:
                cx = kwargs.pop('context', group)
                if group:
                    kwargs['group'] = group
                return self.new_flow(name, ref_quantity=reference, external_ref=external_ref, context=cx, **kwargs)
            except UnknownRefQuantity:
                # assume reference is a unit string specification
                return self.new_quantity(name, ref_unit=reference, external_ref=external_ref, group=group, **kwargs)

    def new_flow(self, name, ref_quantity=None, **kwargs):
        """

        :param name:
        :param ref_quantity: defaults to "Number of items"
        :param kwargs:
        :return:
        """

        if ref_quantity is None:
            ref_quantity = 'Number of items'
        try:
            ref_q = self.get_canonical(ref_quantity)
        except EntityNotFound:
            raise UnknownRefQuantity(ref_quantity)
        f = new_flow(name, ref_q, **kwargs)
        self._archive.add_entity_and_children(f)
        return self.get(f.link)

    def find_term(self, term_ref, origin=None, **kwargs):
        """

        :param term_ref:
        :param origin:
        :param kwargs:
        :return:
        """
        logging.warning('DEPRECATED: find_term()')
        if term_ref is None:
            return
        if hasattr(term_ref, 'entity_type'):
            if term_ref.entity_type == 'context':
                found_ref = term_ref
            elif (not term_ref.is_entity) and (not term_ref.resolved):  # unresolved catalog ref
                try:
                    found_ref = self.get(term_ref.external_ref)
                except EntityNotFound:
                    found_ref = term_ref  # why would we sub an unresolved catalog ref FOR an unresolved catalog ref?
            else:
                found_ref = term_ref
        else:
            # first context
            cx = self._archive.tm[term_ref]
            if cx not in (None, NullContext):
                found_ref = cx
            else:
                found_ref = self.get_local('/'.join(filter(None, (origin, term_ref))))
                ''' # this is now internal to get()
                except EntityNotFound:
                    if origin is None:
                        try:
                            origin, external_ref = term_ref.split('/', maxsplit=1)
                        except ValueError:
                            origin = 'foreground'
                            external_ref = term_ref

                        found_ref = self._archive.catalog_ref(origin, external_ref)
                    else:
                        found_ref = self._archive.catalog_ref(origin, term_ref)
                '''

        if found_ref.entity_type in ('flow', 'process', 'fragment', 'context'):
            return found_ref
        raise TypeError('Invalid entity type for termination: %s' % found_ref.entity_type)

    def post_entity_refs(self, entity_refs, **kwargs):
        """
        Not even sure this function is properly designed. do I *have* to construct a model just to post an existing
        entity to an existing foreground?
        :param entity_refs:
        :param kwargs:
        :return:
        """
        for ref in entity_refs:
            r = self._archive.catalog_ref(ref.origin, ref.external_ref, entity_type=ref.entity_type)
            if hasattr(ref, 'properties'):
                for k in ref.properties():
                    r[k] = ref[k]
            self._archive.add(r)

    def new_fragment(self, flow, direction, external_ref=None, **kwargs):
        """
        :param flow:
        :param direction:
        :param external_ref: if provided, observe and name the fragment after creation
        :param kwargs: uuid=None, parent=None, comment=None, value=None, units=None, balance=False;
          **kwargs passed to LcFragment
        :return:
        """
        if isinstance(flow, str):
            flow = self.get(flow)
        if flow.entity_type != 'flow':
            raise TypeError('%s: Not a %s' % (flow, 'flow'))
        frag = create_fragment(flow, direction, origin=self.origin, **kwargs)
        self._archive.add_entity_and_children(frag)
        if external_ref is not None:
            self.observe(frag, name=external_ref)
        return frag

    '''
    This is officially deprecated. let it break.
    def name_fragment(self, fragment, name, auto=None, force=None, **kwargs):
        return self._archive.name_fragment(fragment, name, auto=auto, force=force)
    '''

    def observe(self, fragment, exchange_value=None, units=None, scenario=None, anchor=None,
                anchor_node=None, anchor_flow=None,
                descend=None, name=None, auto=None, force=None,
                accept_all=None, termination=None, term_flow=None):
        """
        All-purpose method to manipulate fragments.
        :param fragment:
        :param exchange_value: default second positional param; exchange value being observed
        :param units: optional, modifies exchange value
        :param scenario: applies to exchange value and termination equially
        :param anchor: must be an Anchor or FlowTermination
        :param anchor_node: anchor target (node or context)
        :param anchor_flow: convert to flow on termination
        :param descend: set on anchor
        :param name: may not be used if a scenario is also supplied
        :param auto: auto-rename on name collision
        :param force: steal name on name collision
        :param accept_all: not allowed; caught and rejected
        :param termination: deprecated. assigned to anchor.
        :param term_flow: deprecated. assigned to anchor
        :param descend: deprecated, assigned to anchor
        :return:
        """
        anchor_target = None
        if anchor:
            if descend is None:
                descend = anchor.descend
            if isinstance(anchor, Anchor):
                if anchor.type == 'node':
                    anchor_target = anchor.node
                    if anchor_flow is None:
                        anchor_flow = anchor.anchor_flow.entity_id
                else:
                    anchor_target = self.get_context(anchor.context)
            else:
                if anchor.term_node:
                    anchor_target = anchor.term_node
                    if anchor_flow is None:
                        anchor_flow = anchor.term_flow

        if anchor_node:  # override
            anchor_target = anchor_node
        elif termination:
            anchor_target = termination
        if term_flow:
            anchor_flow = term_flow
        if accept_all is not None:
            print('%s: cannot "accept all"' % fragment)
        if name is not None:
            if scenario is None:  #
                if fragment.external_ref != name:
                    self._archive.name_fragment(fragment, name, auto=auto, force=force)
                    print('Naming fragment %s -> %s' % (fragment.external_ref, name))
                else:
                    # nothing to do
                    pass
            else:
                print('Ignoring fragment name under a scenario specification')
        if fragment.observable(scenario):
            self._archive.observe_ev(fragment, scenario=scenario, value=exchange_value, units=units)

        else:
            if exchange_value is not None:
                print('Note: Ignoring exchange value %g for unobservable fragment %s [%s]' % (exchange_value,
                                                                                              fragment.external_ref,
                                                                                              scenario))

        if anchor_target is not None:
            self._archive.observe_anchor(fragment, scenario, anchor_target, anchor_flow, descend=descend)

        return fragment.link

    def observe_unit_score(self, fragment, quantity, score, scenario=None, **kwargs):
        """

        :param fragment:
        :param quantity:
        :param score:
        :param scenario:
        :param kwargs:
        :return:
        """
        term = fragment.termination(scenario)
        term.add_lcia_score(quantity, score, scenario=scenario)

    def scenarios(self, fragment, recurse=True, **kwargs):
        if isinstance(fragment, str):
            fragment = self.get(fragment)

        for s in fragment.scenarios(recurse=recurse):
            yield s

    def nodes(self, origin=None, group='StageName', scenario=None, **kwargs):
        for f in self._archive.entities_by_type('fragment'):
            grp = f.get(group, '')
            for sc, t in f.terminations():
                if scenario is not None:
                    if sc != scenario:
                        continue
                if t.is_null or t.is_context:
                    continue
                if (origin is None and t.term_node.origin != self.origin) or t.term_node.origin == origin:
                    ev = f.exchange_value(sc)
                    yield FragmentBranch(f, t, grp, scenario=sc, magnitude=ev, is_cutoff=False)

    def knobs(self, search=None, param_dict=False, **kwargs):
        args = tuple(filter(None, [search]))
        for k in sorted(self._archive.fragments(*args, show_all=True), key=lambda x: x.external_ref):
            if k.is_balance:
                continue
            if k.is_reference:
                continue
            if k.external_ref == k.uuid:  # only generate named fragments
                continue
            if param_dict:
                yield k.parameters()
            else:
                yield k

    def fragments_with_flow(self, flow, direction=None, reference=True, background=None, **kwargs):
        """
        Requires flow identity
        :param flow:
        :param direction:
        :param reference: {True} | False | None
        :param background:
        :param kwargs:
        :return:
        """
        flow = self[flow]  # retrieve by external ref
        for f in self._archive.fragments_with_flow(flow):
            if background is not None:
                if f.is_background != background:
                    continue
            if direction is not None:
                if f.direction != direction:
                    continue
            if reference is False and f.parent is None:
                continue
            if reference and f.parent:
                continue
            yield f

    def clone_fragment(self, frag, tag=None, **kwargs):
        """

        :param frag: the fragment (and subfragments) to clone
        :param tag: string to attach to named external refs
        :param kwargs: passed to new fragment
        :return:
        """
        clone = clone_fragment(frag, tag=tag, **kwargs)
        self._archive.add_entity_and_children(clone)
        return clone

    def split_subfragment(self, fragment, replacement=None, descend=False, **kwargs):
        """
        Given a non-reference fragment, split it off into a new reference fragment, and create a surrogate child
        that terminates to it.

        without replacement:
        Old:   ...parent-->fragment
        New:   ...parent-->surrogate#fragment;   (fragment)

        with replacement:
        Old:   ...parent-->fragment;  (replacement)
        New:   ...parent-->surrogate#replacement;  (fragment);  (replacement)

        :param fragment:
        :param replacement: [None] if non-None, the surrogate is terminated to the replacement instead of the fork.
        :param descend: [False] on new term
        :return:
        """
        if fragment.reference_entity is None:
            raise AttributeError('Fragment is already a reference fragment')
        if replacement is not None:
            if replacement.reference_entity is not None:
                raise InvalidParentChild('Replacement is not a reference fragment')

        surrogate = _fork_fragment(fragment, comment='New subfragment')
        self._archive.add(surrogate)

        fragment.unset_parent()
        if replacement is None:
            surrogate.terminate(fragment, descend=descend)
        else:
            surrogate.terminate(replacement, descend=descend)

        return fragment

    def interpose(self, fragment, balance=True):
        inter = interpose(fragment)
        self._archive.add(inter)
        if balance:
            fragment.set_balance_flow()

        return inter

    def delete_fragment(self, fragment, **kwargs):
        """
        Remove the fragment and all its subfragments from the archive (they remain in memory)
        This does absolutely no safety checking.
        :param fragment:
        :return:
        """
        if isinstance(fragment, str):
            try:
                fragment = self.get(fragment)
            except EntityNotFound:
                return False
        self._archive.delete_fragment(fragment)
        for c in fragment.child_flows:
            self.delete_fragment(c)
        return True

    def save(self, release: ForegroundRelease = None, description=None, author=None, notes=None, major=False,
             save_unit_scores=None):
        if release is None:
            release = ForegroundRelease(major=major, description=description, author=author, notes=notes)
        self._archive.update_metadata(release)
        return self._archive.save(save_unit_scores=save_unit_scores)

    def tree(self, fragment, **kwargs):
        """

        :param fragment:
        :param kwargs:
        :return:
        """
        frag = self._archive.retrieve_or_fetch_entity(fragment)
        return frag.tree()

    def traverse(self, fragment, scenario=None, **kwargs):
        frag = self._archive.retrieve_or_fetch_entity(fragment)
        return frag.traverse(scenario, observed=True)

    def activity(self, fragment, scenario=None, **kwargs):
        top = self.top(fragment)
        if isinstance(top, LcFragment):
            return [f for f in top.traverse(scenario=scenario, **kwargs) if isinstance(f, LcFragment) and
                    f.top() is top]
        else:
            return top.activity(scenario=scenario, **kwargs)

    def clear_unit_scores(self, lcia_method=None):
        self._archive.clear_unit_scores(lcia_method)

    def clear_scenarios(self, terminations=True):
        for f in self._archive.entities_by_type('fragment'):
            f.clear_scenarios(terminations=terminations)

    def fragment_lcia(self, fragment, quantity_ref, scenario=None, refresh=False, mode=None, **kwargs):
        frag = self._archive.retrieve_or_fetch_entity(fragment)
        res = frag.top().fragment_lcia(quantity_ref, scenario=scenario, refresh=refresh, **kwargs)
        if mode == 'flat':  # 'detailed' doesn't make any sense when run locally
            return res.flatten()
        elif mode == 'stage':
            return res.aggregate()
        elif mode == 'anchor':
            return res.terminal_nodes()
        return res

    def create_process_model(self, process, ref_flow=None, set_background=None, StageName=None, **kwargs):
        """
        Create a fragment from the designated process model.  Creates a reference fragment whose exchange value
        is the reference value of the target, linked to a balance flow that is actually anchored to the target.
        Negative directions (i.e. ecoinvent's negative-valued treatment processes) are inverted and presented as
        having their natural direction.
        :param process:
        :param ref_flow:
        :param set_background:
        :param StageName: applied to both the reference and child
        :param kwargs: applied to the reference
        :return:
        """
        rx = process.reference(ref_flow)
        rv = process.reference_value(ref_flow)
        if rv < 0:  # put in to handle Ecoinvent treatment processes
            dirn = comp_dir(rx.direction)
            rv = abs(rv)
        else:
            dirn = rx.direction
        frag = self.new_fragment(rx.flow, dirn, value=rv, observe=True, **kwargs)
        node = self.new_fragment(rx.flow, frag.direction, parent=frag, balance=True)
        if StageName:
            frag['StageName'] = StageName
            node['StageName'] = StageName
        node.terminate(process, term_flow=rx.flow)
        # if set_background:
        #     frag.set_background()
        # self.fragment_from_exchanges(process.inventory(rx), parent=frag,
        #                              include_context=include_context, multi_flow=multi_flow)
        return frag

    def extend_process(self, fragment, scenario=None, include_context=False, **kwargs):
        if fragment.termination(scenario).is_fg and fragment.balance_flow:
            parent = fragment.balance_flow
        else:
            parent = fragment
        term = parent.termination(scenario)
        try:
            term.term_node.check_bg()
            if include_context:
                inv = chain(term.term_node.dependencies(ref_flow=term.term_flow),
                            term.term_node.cutoffs(ref_flow=term.term_flow),
                            term.term_node.emissions(ref_flow=term.term_flow))
            else:
                inv = chain(term.term_node.dependencies(ref_flow=term.term_flow),
                            term.term_node.cutoffs(ref_flow=term.term_flow))
        except BackgroundRequired:
            inv = term.term_node.inventory(term.term_flow)

        self.fragment_from_exchanges(inv, parent=parent,
                                     scenario=scenario,
                                     include_context=include_context,
                                     **kwargs)
        return fragment

    '''
    def extend_process_model(self, fragment, include_elementary=False, terminate=True, **kwargs):
        """
        "Build out" a fragment, creating child flows from its terminal node's intermediate flows
        :param fragment:
        :param include_elementary:
        :param terminate:
        :param kwargs:
        :return:
        """
        fragment.unset_background()
        process = fragment.term.term_node
        rx = fragment.term.term_flow
        for ex in process.inventory(rx):
            if not include_elementary:
                if ex.type in ('context', 'elementary'):
                    continue
            ch = self.new_fragment(ex.flow, ex.direction, value=ex.value, parent=fragment)
            ch.set_background()
            if ex.type in ('cutoff', 'self'):
                continue
            if terminate:
                ch.terminate(self._archive.catalog_ref(process.origin, ex.termination, entity_type='process'),
                             term_flow=ex.flow)
        fragment.observe(accept_all=True)
        return fragment
    '''
    def _grounded_entity(self, entity, **kwargs):
        if (not entity.is_entity) and (not entity.resolved):
            return self._archive.catalog_ref(entity.origin, entity.external_ref, **kwargs)
        else:
            return entity

    '''# Create or update a fragment from a list of exchanges.

    This needs to be an interface method.
    '''
    def fragment_from_exchanges(self, _xg, parent=None, ref=None, scenario=None,
                                term_dict=None,
                                set_background=None,
                                include_context=True):
        """
        If parent is None, first generated exchange is reference flow; and subsequent exchanges are children.
        Else, all generated exchanges are children of the given parent, and if a child flow exists, update it.
        The generated exchanges are assumed to be ordered, and are matched to child flows in the order originally created.

        This is all tricky if we expect it to work with both ExchangeRefs and actual exchanges (which, obviously, we
        should) because: ExchangeRefs may have only a string for process and flow, but Exchanges will have entities
        for each.  We need to get(flow_ref) and find_term(term_ref) but we can simply use flow entities and catalog
        refs if we have process.origin.  For now we will just swiss-army-knife it.

        And now we are adding auto-terminate to anything that comes back from fragments_with_flow

        :param _xg: Generates a list of exchanges or exchange references
        :param parent: if None, create parent from first exchange. If parent is provided, _xg must exclude references
        :param ref: if parent is created, assign it a name
        :param scenario: [None] specify the scenario under which to terminate child flows
        :param term_dict: [None] a mapping from EITHER existing termination OR flow external ref to target OR (target, term_flow) tuple
        :param set_background: [None] DEPRECATED / background is meaningless
        :param include_context: [False] whether to model context-terminated flows as child fragments
        :return:
        """
        if term_dict is None:
            term_dict = {}

        if set_background is not None:
            print('Warning: set_background is no longer meaningful- all terminations are background')
        if parent is None:
            try:
                x = next(_xg)
            except TypeError:
                x = _xg.pop(0)
            if ref is not None:
                parent = self[ref]

            if parent is None:
                parent = self.new_fragment(self._grounded_entity(x.flow), x.direction, value=x.value, units=x.unit,
                                           Name=str(x.process), **x.args)
                print('Creating new fragment %s (%s)' % (x.process.name, parent.uuid))
                if ref is not None:
                    self.observe(parent, name=ref)

            else:
                self.observe(parent, exchange_value=x.value, units=x.unit)

        _children = list(parent.child_flows)

        for y in _xg:
            """
            Determine flow specification
            """
            if hasattr(y.flow, 'entity_type') and y.flow.entity_type == 'flow':
                try:
                    flow = self._grounded_entity(y.flow)
                except EntityNotFound:
                    flow = y.flow  # groundless flow, better than throwing an exception
            else:
                flow = self[y.flow]
                if flow is None:
                    print('Skipping unknown flow %s' % y.flow)
                    continue
            """
            Determine / retrieve termination
            """
            if y.termination in term_dict:
                term = term_dict[y.termination]
            elif y.flow.external_ref in term_dict:
                term = term_dict[y.flow.external_ref]
            else:
                if y.type == 'context':
                    term = y.termination
                elif y.type == 'self':
                    term = None  # cutoff self-termination
                else:
                    try:
                        if hasattr(y.process, 'origin'):
                            term = self._archive.catalog_ref(y.process.origin, y.termination)
                        else:
                            term = self.get(y.termination)
                    except EntityNotFound:
                        term = None
            if isinstance(term, tuple):
                term, term_flow = term
            else:
                term_flow = None

            if term is not None and term.entity_type == 'context':
                if include_context is False:
                    continue
            elif term == y.process:
                # TODO: figure out why tuple(CatalogRef()) hangs
                term = None  # don't terminate self-term

            """
            Try and match the flow+direction spec to the ordered list of child flows 
            """
            if _children:
                try:
                    c_up = next(g for g in _children if g.flow == flow and g.direction == y.direction)

                    '''
                    #
                    # TODO: children_with_flow needs to be scenario aware
                    # TODO: Update fails with multi_flow when terms are not specified- bc there is no way to tell which
                    # record corresponds to which child.
                    if multi_flow:
                        c_up = next(parent.children_with_flow(flow, direction=y.direction, termination=term,
                                                              recurse=False))
                    else:
                        c_up = next(parent.children_with_flow(flow, direction=y.direction, recurse=False))
                    '''

                    # update value
                    v = y.value
                    if y.unit is not None:
                        v *= c_up.flow.reference_entity.convert(y.unit)

                    if c_up.exchange_value(scenario) != v:
                        print('Updating %s exchange value %.3f %s' % (c_up, y.value, y.unit or ''))

                        self.observe(c_up, exchange_value=y.value, units=y.unit, scenario=scenario)

                    '''# we certainly can
                    if multi_flow:
                        continue  # cannot update terms in the multi-flow case
                    '''
                    '''# However: we should NOT update already-terminated flows with "first available"
                    The current approach:
                     - if no termination is specified, hunt for one
                     - if a termination is supplied OR found in a hunt, go forward:
                       - if the termination doesn't match the existing one, replace it! very destructive
                    
                    RESOLVED: we should NOT hunt for terminations on a fragment update. REASON:
                     * fragments are built in an order selected by the modeler so as to determine which frags are found
                       in a "hunt". If we start updating based on hunt results, we can terminate intentionally-cutoff
                       frags.
                    The proposed new workflow:
                     = if a termination is specified:
                       - if it differs from the existing termination:
                         replace it!
                       - else, nothing to do
                     - else, do nothing. don't go hunting
                    '''
                    '''
                    if term is None:  
                        try:
                            term = next(self.fragments_with_flow(c_up.flow, c_up.direction))
                        except StopIteration:
                            pass
                    '''

                    # set term
                    if term is not None:
                        if term != c_up.term.term_node:
                            print('Updating %s termination %s' % (c_up, term))
                            c_up.clear_termination(scenario)
                            c_up.terminate(term, scenario=scenario, term_flow=term_flow, descend=False)  # none unless specified
                            '''
                            if term.entity_type == 'process' and set_background:
                                c_up.set_background()
                            '''

                    """
                    Propagate properties
                    """
                    for k, v in y.args.items():
                        c_up[k] = v

                    _children.remove(c_up)
                    continue

                except StopIteration:
                    print('No child flow found; creating new %s %s' % (flow, y.direction))
                    pass

            c = self.new_fragment(flow, y.direction, value=y.value, units=y.unit, parent=parent, **y.args)

            if term is None:
                try:  # go hunting for a term in the local foreground
                    term = next(self.fragments_with_flow(c.flow, c.direction))
                except StopIteration:
                    pass

            if term is not None:
                c.terminate(term, scenario=scenario, term_flow=term_flow, descend=False)  # already sets stage name
                '''
                if term.entity_type in ('process', 'flow'):  ##?? whaaa? process should terminate to reference, flow should terminate to itself
                    c.terminate(term, scenario=scenario, term_flow=c.flow)
                    if set_background:
                        c.set_background()
                else:
                    c.terminate(term, scenario=scenario)
                '''
            self.observe(c)  # use cached implicitly via fg interface

        return parent

    '''
    def make_fragment_trees(self, exchanges):
        """
        Take in a list of exchanges [that are properly connected] and build fragment trees from them. Return all roots.

        If an exchange's process has been encountered, it will be used as the parent.  Exchanges with null terminations
        become cutoff flows.

        If an exchange's process has not been encountered, AND its termination is null, it will become a reference
        fragment and be terminated to the process.

        Non-null-terminated exchanges whose processes have not been encountered cause an error.

        This function will only generate new fragments and will not affect any existing fragments.
        :param exchanges:
        :return:
        """
        roots = []
        parents = dict()
        for x in exchanges:
            if x.process.external_ref in parents:
                parent = parents[x.process.external_ref]
                # parent.unset_background()
                frag = self.new_fragment(x.flow, x.direction, parent=parent, **x.args)
                term = self.find_term(x.termination, origin=x.process.origin)
                if term is not None:
                    frag.terminate(term)
                    # frag.set_background()
            else:
                # unmatched flows become roots
                if x.termination is not None:
                    raise InvalidParentChild('Reference flow may not be terminated')
                frag = self.new_fragment(x.flow, x.direction, **x.args)
                roots.append(frag)
                frag.terminate(x.process)

            self.observe(frag, exchange_value=x.value)
            if frag.term.is_process:
                parents[frag.term.term_ref] = frag

        for r in roots:
            yield r
    '''
